import logging
import os
from argparse import ArgumentDefaultsHelpFormatter
from datetime import datetime, timedelta
from statistics import mean

from sbws.globals import fail_hard
from sbws.lib.resultdump import (
    Result,
    ResultError,
    ResultErrorCircuit,
    ResultErrorStream,
    ResultSuccess,
    load_recent_results_in_datadir,
)

log = logging.getLogger(__name__)


def _print_stats_error_types(data):
    counts = {"total": 0}
    for fp in data:
        results = data[fp]
        for result in results:
            if result.type not in counts:
                log.debug("Found a %s for the first time", result.type)
                counts[result.type] = 0
            counts[result.type] += 1
            counts["total"] += 1
    for count_type in counts:
        if count_type == "total":
            continue
        if "error" not in count_type:
            continue
        number = counts[count_type]
        print(
            "{}/{} ({:.2f}%) results were {}".format(
                number,
                counts["total"],
                100 * number / counts["total"],
                count_type,
            )
        )


def _result_type_per_relay(data, result_type):
    out = {}
    for fp in data:
        out[fp] = len([r for r in data[fp] if isinstance(r, result_type)])
    return out


def _get_box_plot_values(iterable):
    """Return the min, q1, med, q1, and max of the input list or iterable.
    This function is NOT perfect, and I think that's fine for basic statistical
    needs. Instead of median, it will return low or high median. Same for q1
    and q3."""
    if not isinstance(iterable, list):
        iterable = list(iterable)
    iterable.sort()
    length = len(iterable)
    median_idx = round(length / 2)
    q1_idx = round(length / 4)
    q3_idx = median_idx + q1_idx
    return [
        iterable[0],
        iterable[q1_idx],
        iterable[median_idx],
        iterable[q3_idx],
        iterable[length - 1],
    ]


def _print_results_type_box_plot(data, result_type):
    per_relay = _result_type_per_relay(data, result_type)
    bp = _get_box_plot_values(per_relay.values())
    print(
        "For {}: min={} q1={} med={} q3={} max={}".format(
            result_type.__name__, *bp
        )
    )


def _print_averages(data):
    mean_success = mean(
        [
            len([r for r in data[fp] if isinstance(r, ResultSuccess)])
            for fp in data
        ]
    )
    print(
        "Mean {:.2f} successful measurements per " "relay".format(mean_success)
    )
    _print_results_type_box_plot(data, Result)
    _print_results_type_box_plot(data, ResultSuccess)
    _print_results_type_box_plot(data, ResultErrorCircuit)
    _print_results_type_box_plot(data, ResultErrorStream)


def _results_into_bandwidths(results, limit=5):
    """
    For all the given resutls, extract their download statistics and normalize
    them into bytes/second bandwidths.

    :param list results: list of :class:`sbws.list.resultdump.ResultSuccess`
    :param int limit: The maximum number of bandwidths to return
    :returns: list of up to `limit` bandwidths, with the largest first
    """
    downloads = []
    for result in results:
        assert isinstance(result, ResultSuccess)
        for dl in result.downloads:
            downloads.append(dl["amount"] / dl["duration"])
    return sorted(downloads, reverse=True)[:limit]


def print_stats(args, data):
    """
    Called from main to print various statistics about the organized **data**
    to stdout.

    :param argparse.Namespace args: command line arguments
    :param dict data: keyed by relay fingerprint, and with values of
        :class:`sbws.lib.resultdump.Result` subclasses
    """
    results = []
    for fp in data:
        results.extend(data[fp])
    assert len([r for r in results if not isinstance(r, Result)]) == 0
    error_results = [r for r in results if isinstance(r, ResultError)]
    success_results = [r for r in results if isinstance(r, ResultSuccess)]
    percent_success_results = 100 * len(success_results) / len(results)
    fastest_transfers = _results_into_bandwidths(success_results)
    fastest_transfer = (
        0 if len(fastest_transfers) < 1 else fastest_transfers[0]
    )
    first_time = min([r.time for r in results])
    last_time = max([r.time for r in results])
    first = datetime.utcfromtimestamp(first_time)
    first = first - timedelta(microseconds=first.microsecond)
    last = datetime.utcfromtimestamp(last_time)
    last = last - timedelta(microseconds=last.microsecond)
    duration = last - first
    print(len(data), "relays have recent results")
    _print_averages(data)
    print(
        len(results),
        "total results, and {:.1f}% are successes".format(
            percent_success_results
        ),
    )
    print(
        len(success_results),
        "success results and",
        len(error_results),
        "error results",
    )
    print(
        "The fastest download was {:.2f} KiB/s".format(fastest_transfer / 1024)
    )
    print("Results come from", first, "to", last, "over a period of", duration)
    if getattr(args, "error_types", False) is True:
        _print_stats_error_types(data)


def gen_parser(sub):
    """
    Helper function for the broader argument parser generating code that adds
    in all the possible command line arguments for the stats command.

    :param argparse._SubParsersAction sub: what to add a sub-parser to
    """
    d = "Write some statistics about the data collected so far to stdout"
    p = sub.add_parser(
        "stats", formatter_class=ArgumentDefaultsHelpFormatter, description=d
    )
    p.add_argument(
        "--error-types",
        action="store_true",
        help="Also print information about each error type",
    )


def main(args, conf):
    """
    Main entry point into the stats command.

    :param argparse.Namespace args: command line arguments
    :param configparser.ConfigParser conf: parsed config files
    """

    datadir = conf.getpath("paths", "datadir")
    if not os.path.isdir(datadir):
        fail_hard("%s does not exist", datadir)

    fresh_days = conf.getint("general", "data_period")
    results = load_recent_results_in_datadir(
        fresh_days, datadir, success_only=False
    )
    if len(results) < 1:
        log.warning("No fresh results")
        return
    print_stats(args, results)



"""
Classes and functions to implement a heartbeat system to monitor the progress.
"""
import logging
import time

from ..util.state import State

log = logging.getLogger(__name__)


class Heartbeat(object):
    """
    Tracks current status of sbws and is capable of printing periodic
    information about the current state
    """

    def __init__(self, state_path):
        # Variable to count total progress in the last days:
        # In case it is needed to see which relays are not being measured,
        # store their fingerprint, not only their number.
        self.measured_fp_set = set()
        self.consensus_fp_set = set()
        self.measured_percent = 0
        self.main_loop_tstart = time.monotonic()

        self.state_dict = State(state_path)

        self.previous_measurement_percent = 0

    def register_measured_fpr(self, async_result):
        self.measured_fp_set.add(async_result)

    def register_consensus_fprs(self, relay_fprs):
        for r in relay_fprs:
            self.consensus_fp_set.add(r)

    def print_heartbeat_message(self):
        """Print the new percentage of the different relays that were measured.

        This way it can be known whether the scanner is making progress
        measuring all the Network.

        Log the percentage, the number of relays measured and not measured,
        the number of loops and the time elapsed since it started measuring.
        """
        loops_count = self.state_dict.count("recent_priority_list")

        not_measured_fp_set = self.consensus_fp_set.difference(
            self.measured_fp_set
        )
        main_loop_tdelta = (time.monotonic() - self.main_loop_tstart) / 60
        new_measured_percent = round(
            len(self.measured_fp_set) / len(self.consensus_fp_set) * 100
        )

        log.info("Run %s main loops.", loops_count)
        log.info(
            "Measured in total %s (%s%% of the last consensus) unique relays"
            " in %i minutes",
            len(self.measured_fp_set),
            new_measured_percent,
            main_loop_tdelta,
        )
        log.info(
            "%s relays from the last consensus are not measured yet.",
            len(not_measured_fp_set),
        )

        # The case when it is equal will only happen when all the relays
        # have been measured.
        if new_measured_percent <= self.previous_measurement_percent:
            log.warning(
                "There is no progress measuring new unique relays with respect"
                " to the previous loop."
            )

        self.previous_measurement_percent = new_measured_percent



import copy
import logging
import os
from threading import RLock

import socks
import stem.process
from stem import (
    ControllerError,
    InvalidArguments,
    InvalidRequest,
    OperationFailed,
    ProtocolError,
    SocketClosed,
    SocketError,
    UnsatisfiableRequest,
)
from stem.connection import IncorrectSocketType
from stem.control import Controller, Listener

from sbws import settings
from sbws.globals import (
    TORRC_OPTIONS_CAN_FAIL,
    TORRC_RUNTIME_OPTIONS,
    TORRC_STARTING_POINT,
    fail_hard,
)

log = logging.getLogger(__name__)
stream_building_lock = RLock()


def attach_stream_to_circuit_listener(controller, circ_id):
    """Returns a function that should be given to add_event_listener(). It
    looks for newly created streams and attaches them to the given circ_id"""

    def closure_stream_event_listener(st):
        if st.status == "NEW" and st.purpose == "USER":
            log.debug(
                "Attaching stream %s to circ %s %s",
                st.id,
                circ_id,
                circuit_str(controller, circ_id),
            )
            try:
                controller.attach_stream(st.id, circ_id)
            # So far we never saw this error.
            except (
                UnsatisfiableRequest,
                InvalidRequest,
                OperationFailed,
            ) as e:
                log.debug(
                    "Error attaching stream %s to circ %s: %s",
                    st.id,
                    circ_id,
                    e,
                )
        else:
            pass

    return closure_stream_event_listener


def add_event_listener(controller, func, event):
    try:
        controller.add_event_listener(func, event)
    except ProtocolError as e:
        log.exception("Exception trying to add event listener %s", e)


def remove_event_listener(controller, func):
    try:
        controller.remove_event_listener(func)
    except SocketClosed as e:
        if not settings.end_event.is_set():
            log.debug(e)
        else:
            log.exception(e)
    except ProtocolError as e:
        log.exception("Exception trying to remove event %s", e)


def init_controller(conf):
    c = None
    # If the external control port is set, use it to initialize the controller.
    control_port = conf["tor"]["external_control_port"]
    if control_port:
        control_port = int(control_port)
        # If it can not connect, the program will exit here
        c = _init_controller_port(control_port)
    # There is no configuration for external control socket, therefore do not
    # attempt to connect to the control socket.
    return c


def is_bootstrapped(c):
    try:
        line = c.get_info("status/bootstrap-phase")
    except (ControllerError, InvalidArguments, ProtocolError) as e:
        log.exception("Error trying to check bootstrap phase %s", e)
        return False
    state, _, progress, *_ = line.split()
    progress = int(progress.split("=")[1])
    if state == "NOTICE" and progress == 100:
        return True
    log.debug("Not bootstrapped. state={} progress={}".format(state, progress))
    return False


def _init_controller_port(port):
    assert isinstance(port, int)
    try:
        c = Controller.from_port(port=port)
        c.authenticate()
    except (IncorrectSocketType, SocketError):
        fail_hard("Unable to connect to control port %s.", port)
    # TODO: Allow for auth via more than just CookieAuthentication
    log.info("Connected to tor via port %s", port)
    return c


def _init_controller_socket(socket):
    assert isinstance(socket, str)
    try:
        c = Controller.from_socket_file(path=socket)
        c.authenticate()
    except (IncorrectSocketType, SocketError):
        log.debug("Error initting controller socket: socket error.")
        return None
    except Exception as e:
        log.exception("Error initting controller socket: %s", e)
        return None
    # TODO: Allow for auth via more than just CookieAuthentication
    log.info("Connected to tor via socket %s", socket)
    return c


def parse_user_torrc_config(torrc, torrc_text):
    """Parse the user configuration torrc text call `extra_lines`
    to a dictionary suitable to use with stem and return a new torrc
    dictionary that merges that dictionary with the existing torrc.
    Example::

        [tor]
        extra_lines =
            Log debug file /tmp/tor-debug.log
            NumCPUs 1
    """
    torrc_dict = torrc.copy()
    for line in torrc_text.split("\n"):
        # Remove leading and trailing whitespace, if any
        line = line.strip()
        # Ignore blank lines
        if len(line) < 1:
            continue
        # Some torrc options are only a key, some are a key value pair.
        kv = line.split(None, 1)
        if len(kv) > 1:
            key, value = kv
        else:
            key = kv[0]
            value = None
        # It's really easy to add to the torrc if the key doesn't exist
        if key not in torrc:
            torrc_dict.update({key: value})
        # But if it does, we have to make a list of values. For example, say
        # the user wants to add a SocksPort and we already have
        # 'SocksPort auto' in the torrc. We'll go from
        #     torrc['SocksPort'] == 'auto'
        # to
        #     torrc['SocksPort'] == ['auto', '9050']
        else:
            existing_val = torrc[key]
            if isinstance(existing_val, str):
                torrc_dict.update({key: [existing_val, value]})
            else:
                assert isinstance(existing_val, list)
                existing_val.append(value)
                torrc_dict.update({key: existing_val})
        log.debug(
            'Adding "%s %s" to torrc with which we are launching Tor',
            key,
            value,
        )
    return torrc_dict


def set_torrc_starting_point(controller):
    """Set the torrc starting point options."""
    for k, v in TORRC_STARTING_POINT.items():
        try:
            controller.set_conf(k, v)
        except (ControllerError, InvalidRequest, InvalidArguments) as e:
            log.exception("Error setting option %s, %s: %s", k, v, e)
            exit(1)


def set_torrc_runtime_options(controller):
    """Set torrc options at runtime."""
    try:
        controller.set_options(TORRC_RUNTIME_OPTIONS)
    # Only the first option that fails will be logged here.
    # Just log stem's exceptions.
    except (ControllerError, InvalidRequest, InvalidArguments) as e:
        log.exception(e)
        exit(1)


def set_torrc_options_can_fail(controller):
    """Set options that can fail, at runtime.

    They can be set at launch, but since the may fail because they are not
    supported in some Tor versions, it's easier to try one by one at runtime
    and ignore the ones that fail.
    """
    for k, v in TORRC_OPTIONS_CAN_FAIL.items():
        try:
            controller.set_conf(k, v)
        except (InvalidArguments, InvalidRequest) as error:
            log.debug(
                "Ignoring option not supported by this Tor version. %s", error
            )
        except ControllerError as e:
            log.exception(e)
            exit(1)


def launch_tor(conf):
    os.makedirs(conf.getpath("tor", "datadir"), mode=0o700, exist_ok=True)
    os.makedirs(conf.getpath("tor", "log"), exist_ok=True)
    os.makedirs(conf.getpath("tor", "run_dpath"), mode=0o700, exist_ok=True)
    # Bare minimum things, more or less
    torrc = copy.deepcopy(TORRC_STARTING_POINT)
    # Very important and/or common settings that we don't know until runtime
    # The rest of the settings are in globals.py
    torrc.update(
        {
            "DataDirectory": conf.getpath("tor", "datadir"),
            "PidFile": conf.getpath("tor", "pid"),
            "ControlSocket": conf.getpath("tor", "control_socket"),
            "Log": [
                "NOTICE file {}".format(
                    os.path.join(conf.getpath("tor", "log"), "notice.log")
                ),
            ],
            "CircuitBuildTimeout": conf["general"]["circuit_timeout"],
        }
    )

    torrc = parse_user_torrc_config(torrc, conf["tor"]["extra_lines"])
    # Finally launch Tor
    try:
        # If there is already a tor process running with the same control
        # socket, this will exit here.
        stem.process.launch_tor_with_config(
            torrc, init_msg_handler=log.debug, take_ownership=True
        )
    except Exception as e:
        fail_hard("Error trying to launch tor: %s", e)
    log.info("Started own tor.")
    # And return a controller to it
    cont = _init_controller_socket(conf.getpath("tor", "control_socket"))
    # In the case it was not possible to connect to own tor socket.
    if not cont:
        fail_hard("Could not connect to own tor control socket.")
    return cont


def launch_or_connect_to_tor(conf):
    cont = init_controller(conf)
    if not cont:
        cont = launch_tor(conf)
    else:
        if not is_torrc_starting_point_set(cont):
            set_torrc_starting_point(cont)
    # Set options that can fail at runtime
    set_torrc_options_can_fail(cont)
    # Set runtime options
    set_torrc_runtime_options(cont)
    log.info("Started or connected to Tor %s.", cont.get_version())
    return cont


def get_socks_info(controller):
    """Returns the first SocksPort Tor is configured to listen on, in the form
    of an (address, port) tuple"""
    try:
        socks_ports = controller.get_listeners(Listener.SOCKS)
        return socks_ports[0]
    except SocketClosed as e:
        if not settings.end_event.is_set():
            log.debug(e)
    # This might need to return the exception if this happen in other cases
    # than when stopping the scanner.
    except ControllerError as e:
        log.debug(e)


def only_relays_with_bandwidth(relays, min_bw=None, max_bw=None):
    """
    Given a list of relays, only return those that optionally have above
    **min_bw** and optionally have below **max_bw**, inclusively. If neither
    min_bw nor max_bw are given, essentially just returns the input list of
    relays.
    """
    assert min_bw is None or min_bw >= 0
    assert max_bw is None or max_bw >= 0
    ret = []
    for relay in relays:
        assert hasattr(relay, "consensus_bandwidth")
        if min_bw is not None and relay.consensus_bandwidth < min_bw:
            continue
        if max_bw is not None and relay.consensus_bandwidth > max_bw:
            continue
        ret.append(relay)
    return ret


def circuit_str(controller, circ_id):
    assert isinstance(circ_id, str)
    int(circ_id)
    try:
        circ = controller.get_circuit(circ_id)
    except ValueError as e:
        log.warning(
            "Circuit %s no longer seems to exist so can't return "
            "a valid circuit string for it: %s",
            circ_id,
            e,
        )
        return None
    # exceptions raised when stopping the scanner
    except (ControllerError, SocketClosed, socks.GeneralProxyError) as e:
        log.debug(e)
        return None
    return (
        "["
        + " -> ".join(["{} ({})".format(n, fp[0:8]) for fp, n in circ.path])
        + "]"
    )


def is_torrc_starting_point_set(tor_controller):
    """Verify that the tor controller has the correct configuration.

    When connecting to a tor controller that has not been launched by sbws,
    it should have been configured to work with sbws.

    """
    bad_options = False
    torrc = TORRC_STARTING_POINT
    for k, v in torrc.items():
        value_set = tor_controller.get_conf(k)
        if v != value_set:
            log.exception(
                "Incorrectly configured %s, should be %s, is %s",
                k,
                v,
                value_set,
            )
            bad_options = True
    if not bad_options:
        log.info("Tor is correctly configured to work with sbws.")
    return bad_options



import requests

from sbws import settings
from sbws.util import stem as stem_utils


class TimedSession(requests.Session):
    """Requests Session that sends timeout in the head and get methods."""

    def get(self, url, **kwargs):
        return super().get(
            url, timeout=getattr(self, "_timeout", None), **kwargs
        )

    def head(self, url, **kwargs):
        return super().head(
            url, timeout=getattr(self, "_timeout", None), **kwargs
        )


def make_session(controller, timeout):
    """
    Initialize a TimedSession with the timeout, the proxies and the headers.

    """
    s = TimedSession()
    socks_info = stem_utils.get_socks_info(controller)
    # Probably because scanner is stopping.
    if socks_info is None:
        return None
    s.proxies = {
        "http": "socks5h://{}:{}".format(*socks_info),
        "https": "socks5h://{}:{}".format(*socks_info),
    }
    # ``_timeout`` is not used by request's Session, but it is by TimedSession.
    s._timeout = timeout
    s.headers = settings.HTTP_HEADERS
    return s



"""Unit tests for heartbeat"""
import logging

from sbws.lib import heartbeat


def test_register_measured_fpr(conf):
    """
    Test that checks if the values entered into the function
    register_measured_fpr is inputted into the variable measured_fp.set

    Asserts if the value of measured_fp_set is the same value
    inputted into the function register_measured_fpr which is "A"
    """
    hbeat = heartbeat.Heartbeat(conf.getpath("paths", "state_fname"))
    hbeat.register_measured_fpr("A")
    hbeat.register_measured_fpr("B")
    assert "A" in hbeat.measured_fp_set
    assert "B" in hbeat.measured_fp_set


def test_register_consensus_fprs(conf):
    """
    Test that checks if the values entered into the function
    register_consensus_fprs is inputted into the variable consensus_fp_set

    Asserts if the value of consensus_fp_set is the same value
    inputted into the function register_consensus_fprs which is ["4","5","6"]
    """
    hbeat = heartbeat.Heartbeat(conf.getpath("paths", "state_fname"))
    hbeat.register_consensus_fprs(["4", "5", "6"])
    assert hbeat.consensus_fp_set == {"4", "5", "6"}


def test_print_heartbeat_message(conf, caplog):
    """
    Test that checks if the print_message function is working as predicted.
    Also, to check if the text is also logged as expected.

    Asserts that the specified text is present in the log
    """
    hbeat = heartbeat.Heartbeat(conf.getpath("paths", "state_fname"))
    hbeat.register_measured_fpr("12")
    hbeat.register_consensus_fprs(["1", "2", "4"])
    hbeat.print_heartbeat_message()

    assert hbeat.previous_measurement_percent == 33
    assert "Run None main loops." in caplog.records[0].getMessage()
    assert "Measured in total 1 (33%" in caplog.records[1].getMessage()
    log_relay_not_measured = "3 relays from the last consensus are not "
    assert log_relay_not_measured in caplog.records[2].getMessage()

    caplog.clear()
    new_percent = round(
        len(hbeat.measured_fp_set) / len(hbeat.consensus_fp_set) * 100
    )
    hbeat.previous_measurement_percent = new_percent + 100
    caplog.set_level(logging.WARNING)

    hbeat.print_heartbeat_message()
    log_no_progress = "There is no progress measuring new unique relays"
    assert log_no_progress in caplog.records[0].getMessage()



"""Unit tests for scaling.py."""
import os
from statistics import mean

from sbws.lib import scaling
from sbws.lib.resultdump import ResultSuccess, load_result_file


def test_bw_filt():
    bw_measurements = [
        96700.00922329757,
        70311.63051659254,
        45531.743347556374,
        38913.97025485627,
        55656.332364676025,
    ]
    fb = scaling.bw_filt(bw_measurements)
    # This is greater than the mean, that is 61422.73714139576
    assert fb == 83506

    # When there are no measurements what can not be the case for successful
    # results.
    bw_measurements = []
    assert 0 == scaling.bw_filt(bw_measurements)

    bw_measurements = [1, 0]
    # Because rounded to int
    assert 0 == round(mean(bw_measurements))
    # So the filtered bw will be also 0
    assert 0 == scaling.bw_filt(bw_measurements)

    bw_measurements = [1, 2, 3]
    # Because rounded to int
    assert 2 == round(mean(bw_measurements))
    assert 2 == scaling.bw_filt(bw_measurements)

    bw_measurements = [10, 0]
    assert 5 == round(mean(bw_measurements))
    # Because the value 10 is bigger than the mean
    assert 10 == scaling.bw_filt(bw_measurements)

    bw_measurements = [0, 10, 20]
    assert 10 == round(mean(bw_measurements))
    # Because 10 and 20 are bigger or equal than the mean
    assert 15 == scaling.bw_filt(bw_measurements)


def test_bw_filt_from_results(root_data_path):
    results_file = os.path.join(
        root_data_path, ".sbws", "datadir", "2019-03-25.txt"
    )
    results = load_result_file(results_file)
    bw_filts = {}
    for fp, values in results.items():
        success_results = [r for r in values if isinstance(r, ResultSuccess)]
        if success_results:
            bw_measurements = scaling.bw_measurements_from_results(
                success_results
            )
            mu = round(mean(bw_measurements))
            muf = scaling.bw_filt(bw_measurements)
            bw_filts[fp] = (mu, muf)
    for fp, values in bw_filts.items():
        assert bw_filts[fp][0] <= bw_filts[fp][1]
    assert 5526756 == bw_filts["117A456C911114076BEB4E757AC48B16CC0CCC5F"][0]
    assert 5643086 == bw_filts["117A456C911114076BEB4E757AC48B16CC0CCC5F"][1]
    assert 5664965 == bw_filts["693F73187624BE760AAD2A12C5ED89DB1DE044F5"][0]
    assert 5774274 == bw_filts["693F73187624BE760AAD2A12C5ED89DB1DE044F5"][1]
    assert 5508279 == bw_filts["270A861ABED22EC2B625198BCCD7B2B9DBFFC93C"][0]
    assert 5583737 == bw_filts["270A861ABED22EC2B625198BCCD7B2B9DBFFC93C"][1]
    assert 5379911 == bw_filts["E894C65997F8EC96558B554176EEEA39C6A43EF6"][0]
    assert 5485088 == bw_filts["E894C65997F8EC96558B554176EEEA39C6A43EF6"][1]



import logging
import os.path
from unittest.mock import patch

import sbws.core.stats
from tests.unit.globals import monotonic_time


def test_stats_initted(sbwshome_empty, args, conf, caplog):
    """
    An initialized but rather empty .sbws directory should fail about missing
    ~/.sbws/datadir
    """
    try:
        sbws.core.stats.main(args, conf)
    except SystemExit as e:
        assert e.code == 1
    else:
        assert None, "Should have failed"
    assert (
        "{}/datadir does not exist".format(os.path.abspath(sbwshome_empty))
        == caplog.records[-1].getMessage()
    )


def test_stats_stale_result(args, conf, caplog, sbwshome_success_result):
    """
    An initialized .sbws directory with no fresh results should say so and
    exit cleanly
    """
    caplog.set_level(logging.DEBUG)
    sbws.core.stats.main(args, conf)
    assert "No fresh results" == caplog.records[-1].getMessage()


@patch("time.time")
def test_stats_fresh_result(
    time_mock, sbwshome_error_result, args, conf, capsys, caplog
):
    """
    An initialized .sbws directory with a fresh error result should have some
    boring stats and exit cleanly
    """
    args.error_types = False
    start = 1529232278
    time_mock.side_effect = monotonic_time(start=start)
    sbws.core.stats.main(args, conf)
    captured = capsys.readouterr()
    lines = captured.out.strip().split("\n")
    assert "1 relays have recent results" in lines[0]
    # FIXME
    # needed_output_lines = [
    #     '1 relays have recent results',
    #     'Mean 0.00 successful measurements per relay',
    #     '0 success results and 1 error results',
    # ]
    # for needed_line in needed_output_lines:
    #     assert needed_line in lines
    # lines = [l.getMessage() for l in caplog.records]
    # needed_log_lines = [
    #     'Keeping 1/1 read lines from {}/{}/{}.txt'.format(
    #         sbwshome_error_result, 'datadir', '2018-06-17'),
    #     'Keeping 1/1 results after removing old ones',
    # ]
    # for needed_line in needed_log_lines:
    #     assert needed_line in lines


@patch("time.time")
def test_stats_fresh_results(
    time_mock, sbwshome_success_result_two_relays, args, conf, capsys, caplog
):
    """
    An initialized .sbws directory with a fresh error and fresh success should
    have some exciting stats and exit cleanly
    """
    caplog.set_level(logging.DEBUG)
    start = 1529232278
    time_mock.side_effect = monotonic_time(start=start)
    sbws.core.stats.main(args, conf)
    captured = capsys.readouterr()
    lines = captured.out.strip().split("\n")
    assert "1 relays have recent results" in lines[0]
    # FIXME
    # needed_output_lines = [
    #     '1 relays have recent results',
    #     '1 success results and 1 error results',
    #     'Mean 1.00 successful measurements per relay',
    #     '1/2 (50.00%) results were error-misc',
    # ]
    # for needed_line in needed_output_lines:
    #     assert needed_line in lines
    # lines = [l.getMessage() for l in caplog.records]
    # needed_log_lines = [
    #     'Keeping 2/2 read lines from {}/{}/{}.txt'.format(
    #         sbwshome_success_result_two_relays, 'datadir',
    #         datetime.utcfromtimestamp(time.time()).date()),
    #     'Keeping 2/2 results after removing old ones',
    #     'Found a _ResultType.Error for the first time',
    #     'Found a _ResultType.Success for the first time',
    # ]
    # for needed_line in needed_log_lines:
    #     assert needed_line in lines



import json
import os

from sbws.util.filelock import FileLock

from .json import CustomDecoder, CustomEncoder


class State:
    """
    `json` wrapper to read a json file every time it gets a key and to write
    to the file every time a key is set.

    Every time a key is got or set, the file is locked, to atomically access
    and update the file across threads and across processes.

    >>> state = State('foo.state')
    >>> # state == {}

    >>> state['linux'] = True
    >>> # 'foo.state' now exists on disk with the JSON for {'linux': True}

    >>> # We read 'foo.state' from disk in order to get the most up-to-date
    >>> #     state info. Pretend another process has updated 'linux' to be
    >>> #     False
    >>> state['linux']
    >>> # returns False

    >>> # Pretend another process has added the user's age to the state file.
    >>> #     As before, we read the state file from disk for the most
    >>> #     up-to-date info.
    >>> state['age']
    >>> # Returns 14

    >>> # We now set their name. We read the state file first, set the option,
    >>> #     and then write it out.
    >>> state['name'] = 'John'

    >>> # We can do many of the same things with a State object as with a dict
    >>> for key in state: print(key)
    >>> # Prints 'linux', 'age', and 'name'

    """

    def __init__(self, fname):
        self._fname = fname
        self._state = self._read()

    def _read(self):
        if not os.path.exists(self._fname):
            return {}
        with FileLock(self._fname):
            with open(self._fname, "rt") as fd:
                return json.load(fd, cls=CustomDecoder)

    def _write(self):
        with FileLock(self._fname):
            with open(self._fname, "wt") as fd:
                return json.dump(self._state, fd, indent=4, cls=CustomEncoder)

    def __len__(self):
        self._state = self._read()
        return self._state.__len__()

    def get(self, key, d=None):
        """
        Implements a dictionary ``get`` method reading and locking
        a json file.
        """
        self._state = self._read()
        return self._state.get(key, d)

    def __getitem__(self, key):
        self._state = self._read()
        return self._state.__getitem__(key)

    def __delitem__(self, key):
        self._state = self._read()
        self._state.__delitem__(key)
        self._write()

    def __setitem__(self, key, value):
        # NOTE: important, read the file before setting the key,
        # otherwise if other instances are creating other keys, they're lost.
        self._state = self._read()
        self._state.__setitem__(key, value)
        self._write()

    def __iter__(self):
        self._state = self._read()
        return self._state.__iter__()

    def __contains__(self, item):
        self._state = self._read()
        return self._state.__contains__(item)

    def count(self, k):
        """
        Returns the length if the key value is a list
        or the sum of number if the key value is a list of list
        or the key value
        or None if the state doesn't have the key.
        """
        if self.get(k):
            if isinstance(self._state[k], list):
                if isinstance(self._state[k][0], list):
                    return sum(map(lambda x: x[1], self._state[k]))
                return len(self._state[k])
            return self.get(k)
        return None



"""relaylist.py unit tests."""
from datetime import datetime

# When datetime is imported as a class (`from datetime import datetime`) it can
# not be mocked because it is a built-in type. It can only be mocked when
# imported as module.
# freezegun is able to mock any datetime object, it also allows comparisons.
from freezegun import freeze_time

from sbws.lib.relaylist import Relay, RelayList
from sbws.util.state import State


def test_init_relays(
    args, conf, controller, controller_1h_later, controller_5days_later
):
    """
    Test `init_relays` when creating the RelayList the first time and when a
    new consensus is received.
    Test that the number of consesus timesamps and relays is correct.
    Additionally, make sure the calculated min bw for the second hop for
    exit/non-exit relays is correct, too.
    """
    state = State(conf["paths"]["state_fpath"])
    # There is no need to mock datetime to update the consensus, since the
    # actual date will be always later.
    # But it's needed to have the correct list of timestamps both for RelayList
    # and Relay.
    with freeze_time("2020-02-29 10:00:00"):
        relay_list = RelayList(args, conf, controller, state=state)
    assert relay_list.recent_consensus_count == 1
    assert relay_list._relays[0].relay_in_recent_consensus_count == 1
    # The actual number of relays in the consensus
    assert len(relay_list._relays) == 6433
    fps = {r.fingerprint for r in relay_list._relays}
    # The calculated min bw for the second hop
    assert 2100000 == relay_list._exit_min_bw
    assert 220000 == relay_list._non_exit_min_bw

    # One hour later there is a new consensus
    relay_list._controller = controller_1h_later
    with freeze_time("2020-02-29 11:00:00"):
        # Call relays update the list of relays.
        relay_list.relays
    assert relay_list.recent_consensus_count == 2
    assert relay_list._relays[0].relay_in_recent_consensus_count == 2
    # Check that the number of relays is now the previous one plus the relays
    # that are in the new consensus that there were not in the previous one.
    fps_1h_later = {r.fingerprint for r in relay_list._relays}
    added_fps = fps_1h_later.difference(fps)
    assert 6505 == 6433 + len(added_fps)
    # The calculated min bw for the second hop
    assert 2120000 == relay_list._exit_min_bw
    assert 210000 == relay_list._non_exit_min_bw

    # Five days later plus 1 second.
    # The first consensus timestamp will get removed.
    relay_list._controller = controller_5days_later
    with freeze_time("2020-03-05 10:00:01"):
        relay_list.relays
    assert relay_list.recent_consensus_count == 2
    assert relay_list._relays[0].relay_in_recent_consensus_count == 2
    fps_5days_later = {r.fingerprint for r in relay_list._relays}
    # The number of added relays will be the number of relays in this
    # consensus that were not in the other 2 conensuses
    added_fps = fps_5days_later.difference(fps_1h_later)
    # The number of removed relays that are in this consensus, plus the added
    # ones that were not in the first consensus (because it has been removed).
    removed_fps = fps.difference(fps_5days_later)
    # The number of relays will be the number of relays in the cosensus plus
    # the added ones minus the removed ones.
    assert 6596 == 6505 + len(added_fps) - len(removed_fps)
    # The calculated min bw for the second hop
    assert 2800000 == relay_list._exit_min_bw
    assert 150000 == relay_list._non_exit_min_bw


def test_increment_recent_measurement_attempt(args, conf, controller):
    """Test that incrementing the measurement attempts does not go on forever

    And instead it only counts the number of attempts in the last days.
    It also tests that the state file is updated correctly.
    """
    state = State(conf["paths"]["state_fpath"])
    with freeze_time("2020-02-29 10:00:00"):
        relay_list = RelayList(args, conf, controller=controller, state=state)
    # The initial count is 0 and the state does not have that key.
    assert 0 == relay_list.recent_measurement_attempt_count
    assert not state.get("recent_measurement_attempt", None)

    # Pretend that a measurement attempt is made.
    with freeze_time("2020-02-29 10:00:00"):
        relay_list.increment_recent_measurement_attempt()
    assert 1 == relay_list.recent_measurement_attempt_count
    assert [datetime(2020, 2, 29, 10, 0)] == state[
        "recent_measurement_attempt"
    ]

    # And a second measurement attempt is made 4 days later.
    with freeze_time("2020-03-04 10:00:00"):
        relay_list.increment_recent_measurement_attempt()
    assert 2 == relay_list.recent_measurement_attempt_count
    assert 2 == len(state["recent_measurement_attempt"])

    # And a third measurement attempt is made 5 days later.
    with freeze_time("2020-03-05 10:00:00"):
        relay_list.increment_recent_measurement_attempt()
    assert 3 == relay_list.recent_measurement_attempt_count
    assert 3 == len(state["recent_measurement_attempt"])

    # And a fourth measurement attempt is made 6 days later. The first one is
    # now removed and not counted.
    with freeze_time("2020-03-06 10:00:00"):
        relay_list.increment_recent_measurement_attempt()
    assert 3 == relay_list.recent_measurement_attempt_count
    assert 3 == len(state["recent_measurement_attempt"])


def test_increment_relay_recent_measurement_attempt(
    controller, router_status, server_descriptor
):
    """Test that incrementing the measurement attempts do not go on forever

    And instead it only counts the number of attempts in the last days.
    """
    # For this test it does not matter that the consensus timestamps
    # are not correct.
    now = datetime.utcnow()
    relay = Relay("A" * 40, controller, timestamp=now)
    # The initial count is 0 and the state does not have that key.
    assert 0 == relay.relay_recent_measurement_attempt_count

    # Pretend that a measurement attempt is made.
    with freeze_time("2020-02-29 10:00:00"):
        relay.increment_relay_recent_measurement_attempt()
    assert 1 == relay.relay_recent_measurement_attempt_count

    # And a second measurement attempt is made 4 days later.
    with freeze_time("2020-03-04 10:00:00"):
        relay.increment_relay_recent_measurement_attempt()
    assert 2 == relay.relay_recent_measurement_attempt_count

    # And a third measurement attempt is made 5 days later.
    with freeze_time("2020-03-05 10:00:00"):
        relay.increment_relay_recent_measurement_attempt()
    assert 3 == relay.relay_recent_measurement_attempt_count

    # And a fourth measurement attempt is made 6 days later. The first one is
    # now removed and not counted.
    with freeze_time("2020-03-06 10:00:00"):
        relay.increment_relay_recent_measurement_attempt()
    assert 3 == relay.relay_recent_measurement_attempt_count
    # XXX: Write to a Result and load it back


def test_increment_relay_recent_priority_list(
    controller, router_status, server_descriptor
):
    """Test that incrementing the priority lists do not go on forever

    And instead it only counts the number of priority lists in the last days.
    """
    # For this test it does not matter that the consensus timestamps
    # are not correct.
    now = datetime.utcnow()
    relay = Relay("A" * 40, controller, timestamp=now)
    # The initial count is 0 and the state does not have that key.
    assert 0 == relay.relay_recent_priority_list_count

    # Pretend that a measurement attempt is made.
    with freeze_time("2020-02-29 10:00:00"):
        relay.increment_relay_recent_priority_list()
    assert 1 == relay.relay_recent_priority_list_count

    # And a second measurement attempt is made 4 days later.
    with freeze_time("2020-03-04 10:00:00"):
        relay.increment_relay_recent_priority_list()
    assert 2 == relay.relay_recent_priority_list_count

    # And a third measurement attempt is made 5 days later.
    with freeze_time("2020-03-05 10:00:00"):
        relay.increment_relay_recent_priority_list()
    assert 3 == relay.relay_recent_priority_list_count

    # And a fourth measurement attempt is made 6 days later. The first one is
    # now removed and not counted.
    with freeze_time("2020-03-06 10:00:00"):
        relay.increment_relay_recent_priority_list()
    assert 3 == relay.relay_recent_priority_list_count



"""Unit tests for sbws.lib.destination."""
from datetime import datetime, timedelta

from sbws.globals import MAX_SECONDS_RETRY_DESTINATION
from sbws.lib import destination


def test_destination_is_functional():
    eleven_mins_ago = datetime.utcnow() - timedelta(minutes=11)
    six_mins_ago = datetime.utcnow() - timedelta(minutes=6)
    four_mins_ago = datetime.utcnow() - timedelta(minutes=4)
    # Make last time tried a bit bigger than the half of the maximum, so that
    # it's bigger than the delta time to retry, and when delta time to retry
    # is muliplied by a factor (2) it reaches the maximum.
    long_ago = datetime.utcnow() - timedelta(
        (MAX_SECONDS_RETRY_DESTINATION / 2) + 2
    )

    d = destination.Destination("unexistenturl", 0, False)
    assert d.is_functional()

    # Fail 3 consecutive times
    d.add_failure()
    d.add_failure()
    d.add_failure()
    assert d._are_last_attempts_failures()
    assert not d._is_last_try_old_enough()
    assert not d.is_functional()

    # Then doesn't fail and it's functional again
    d.add_success()
    assert not d._are_last_attempts_failures()
    assert d.is_functional()

    # Fail again 3 times
    d.add_failure()
    d.add_failure()
    # And last failure was 2h ago
    d.add_failure(four_mins_ago)
    assert d._are_last_attempts_failures()
    assert not d._is_last_try_old_enough()
    assert not d.is_functional()

    # But if the last failure was 4h ago, try to use it again
    # And last failure was 4h ago
    d.add_failure(six_mins_ago)
    assert d._is_last_try_old_enough()
    assert d.is_functional()

    # If last failure was 8h ago, try to use it again again
    d.add_failure(eleven_mins_ago)
    assert d._is_last_try_old_enough()
    assert d.is_functional()

    # Whenever it does not fail again, reset the time to try again
    # on 3 consecutive failures
    d.add_success()
    assert not d._are_last_attempts_failures()
    assert d.is_functional()
    # And the delta to try is reset
    assert not d._is_last_try_old_enough()

    # When the delta time to retry a destination increase too much,
    # set it to a maximum, and try the destination again
    d.add_failure()
    d.add_failure()
    d.add_failure(long_ago)
    # Pretend the delta seconds was already set to a bit more than
    # half the maximum.
    d._delta_seconds_retry = (MAX_SECONDS_RETRY_DESTINATION / 2) + 1
    assert d._are_last_attempts_failures()
    assert d._is_last_try_old_enough()
    assert d.is_functional()
    assert d._delta_seconds_retry == MAX_SECONDS_RETRY_DESTINATION



def test_relay_properties(rl):
    relay = [relay for relay in rl.relays if relay.nickname == "test000a"][0]
    # The fingerprint and the master key can't be tested cause they are
    # created by chutney.
    assert "Authority" in relay.flags
    assert not relay.exit_policy or not relay.exit_policy.is_exiting_allowed()
    assert relay.average_bandwidth == 1073741824
    # Since tor version 0.4.7.0-alpha-dev, #40337 patch, chutney relays notice
    # bandwidth changes, so consensus bandwidth might be higher than 0.
    assert relay.address == "127.0.0.1"


def test_relay_list_last_consensus_timestamp(rl):
    assert (
        rl.last_consensus_timestamp == rl._relays[0].last_consensus_timestamp
    )



import logging
import os
from argparse import ArgumentDefaultsHelpFormatter
from math import ceil

from sbws.globals import (
    DAY_SECS,
    GENERATE_PERIOD,
    NUM_MIN_RESULTS,
    PROP276_ROUND_DIG,
    SBWS_SCALE_CONSTANT,
    SBWS_SCALING,
    TORFLOW_BW_MARGIN,
    TORFLOW_SCALING,
    fail_hard,
)
from sbws.lib import destination
from sbws.lib.resultdump import load_recent_results_in_datadir
from sbws.lib.v3bwfile import V3BWFile
from sbws.util.timestamp import now_fname

log = logging.getLogger(__name__)


def gen_parser(sub):
    d = (
        "Generate a v3bw file based on recent results. A v3bw file is the "
        "file Tor directory authorities want to read and base their "
        "bandwidth votes on. "
        "To avoid inconsistent reads, configure tor with "
        '"V3BandwidthsFile /path/to/latest.v3bw". '
        "(latest.v3bw is an atomically created symlink in the same "
        "directory as output.) "
        "If the file is transferred to another host, it should be written to "
        "a temporary path, then renamed to the V3BandwidthsFile path.\n"
        "The default scaling method is torflow's one. To use different"
        "scaling methods or no scaling, see the options."
    )
    p = sub.add_parser(
        "generate",
        description=d,
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    p.add_argument(
        "--output",
        default=None,
        type=str,
        help="If specified, write the v3bw here instead of what is"
        "specified in the configuration",
    )
    # The reason for --scale-constant defaulting to 7500 is because at one
    # time, torflow happened to generate output that averaged to 7500 bw units
    # per relay. We wanted the ability to try to be like torflow. See
    # https://lists.torproject.org/pipermail/tor-dev/2018-March/013049.html
    p.add_argument(
        "--scale-constant",
        default=SBWS_SCALE_CONSTANT,
        type=int,
        help="When scaling bw weights, scale them using this const "
        "multiplied by the number of measured relays",
    )
    p.add_argument(
        "--scale-sbws",
        action="store_true",
        help="If specified, do not use bandwidth values as they "
        "are, but scale them such that we have a budget of "
        "scale_constant * num_measured_relays = bandwidth to give "
        "out, and we do so proportionally",
    )
    p.add_argument(
        "-t",
        "--scale-torflow",
        action="store_true",
        default=True,
        help="If specified, scale measurements using torflow's "
        "method. This option is kept for compatibility with older "
        "versions and it is silently ignored, since it is the "
        "default.",
    )
    p.add_argument(
        "-w",
        "--raw",
        action="store_true",
        help="If specified, do use bandwidth raw measurements "
        "without any scaling.",
    )
    p.add_argument(
        "-m",
        "--torflow-bw-margin",
        default=TORFLOW_BW_MARGIN,
        type=float,
        help="Cap maximum bw when scaling as Torflow. ",
    )
    p.add_argument(
        "-r",
        "--round-digs",
        "--torflow-round-digs",
        default=PROP276_ROUND_DIG,
        type=int,
        help="Number of most significant digits to round bw.",
    )
    p.add_argument(
        "-p",
        "--secs-recent",
        default=None,
        type=int,
        help="How many secs in the past are results being "
        "still considered. Default is {} secs. If not scaling "
        "as Torflow the default is data_period in the "
        "configuration.".format(GENERATE_PERIOD),
    )
    p.add_argument(
        "-a",
        "--secs-away",
        default=DAY_SECS,
        type=int,
        help="How many secs results have to be away from each " "other.",
    )
    p.add_argument(
        "-n",
        "--min-num",
        default=NUM_MIN_RESULTS,
        type=int,
        help="Minimum number of a results to consider them.",
    )
    return p


def main(args, conf):
    os.makedirs(conf.getpath("paths", "v3bw_dname"), exist_ok=True)

    datadir = conf.getpath("paths", "datadir")
    if not os.path.isdir(datadir):
        fail_hard("%s does not exist", datadir)
    if args.scale_constant < 1:
        fail_hard("--scale-constant must be positive")
    if args.torflow_bw_margin < 0:
        fail_hard("toflow-bw-margin must be major than 0.")
    if args.scale_sbws:
        scaling_method = SBWS_SCALING
    elif args.raw:
        scaling_method = None
    else:
        # sbws will scale as torflow until we have a better algorithm for
        # scaling (#XXX)
        scaling_method = TORFLOW_SCALING
    if args.secs_recent:
        fresh_days = ceil(args.secs_recent / 24 / 60 / 60)
    elif scaling_method == TORFLOW_SCALING:
        fresh_days = ceil(GENERATE_PERIOD / 24 / 60 / 60)
    else:
        fresh_days = conf.getint("general", "data_period")
    reset_bw_ipv4_changes = conf.getboolean("general", "reset_bw_ipv4_changes")
    reset_bw_ipv6_changes = conf.getboolean("general", "reset_bw_ipv6_changes")
    results = load_recent_results_in_datadir(
        fresh_days,
        datadir,
        on_changed_ipv4=reset_bw_ipv4_changes,
        on_changed_ipv6=reset_bw_ipv6_changes,
    )
    if len(results) < 1:
        log.warning(
            "No recent results, so not generating anything. (Have you "
            "ran sbws scanner recently?)"
        )
        return
    state_fpath = conf.getpath("paths", "state_fname")
    consensus_path = os.path.join(
        conf.getpath("tor", "datadir"), "cached-consensus"
    )
    # Accept None as scanner_country to be compatible with older versions.
    scanner_country = conf["scanner"].get("country")
    destinations_countries = destination.parse_destinations_countries(conf)
    bw_file = V3BWFile.from_results(
        results,
        scanner_country,
        destinations_countries,
        state_fpath,
        args.scale_constant,
        scaling_method,
        torflow_cap=args.torflow_bw_margin,
        round_digs=args.round_digs,
        secs_recent=args.secs_recent,
        secs_away=args.secs_away,
        min_num=args.min_num,
        consensus_path=consensus_path,
    )

    output = args.output or conf.getpath("paths", "v3bw_fname").format(
        now_fname()
    )
    bw_file.write(output)
    bw_file.info_stats



from configparser import ConfigParser

import sbws.util.config as con


class PseudoSection:
    def __init__(self, key, value, mini=None, maxi=None):
        self.key = key
        self.value = value
        self.mini = mini
        self.maxi = maxi

    def getfloat(self, key):
        assert key == self.key, "But in tests; key should exist"
        return float(self.value)

    def getint(self, key):
        assert key == self.key, "But in tests; key should exist"
        return int(self.value)


def test_validate_fingerprint():
    fp_len = 40
    bads = [
        "A" * (fp_len - 1),
        "A" * (fp_len + 1),
        "",
        "A" * (1000000),
        "a" * fp_len,
        "O" * fp_len,
    ]
    goods = ["A" * fp_len, "".join(list("0123456789ABCDEF" * 3)[0:fp_len])]
    for fp in bads:
        d = {"": fp}
        valid, reason = con._validate_fingerprint(d, "")
        assert (
            not valid
        ), "Fingerprint {} should not have passed " "validation".format(fp)
    for fp in goods:
        d = {"": fp}
        valid, reason = con._validate_fingerprint(d, "")
        assert valid, (
            "Fingerprint {} should have passed, but didn't "
            "because {}".format(fp, reason)
        )


def test_validate_int_simple():
    bads = [
        PseudoSection("", "NotAInt"),
        PseudoSection("", "-0.1"),
        PseudoSection("", "0.1"),
    ]
    goods = [
        PseudoSection("", "0"),
        PseudoSection("", "1"),
        PseudoSection("", "-1"),
        PseudoSection("", "100000000"),
        PseudoSection("", "-1000000000"),
        PseudoSection("", "+0"),
        PseudoSection("", "-0"),
    ]
    for sec in bads:
        valid, reason = con._validate_int(sec, sec.key)
        assert not valid, "{} should not have been a valid " "int".format(
            sec.value
        )
    for sec in goods:
        valid, reason = con._validate_int(sec, sec.key)
        assert valid, "{} should have been a valid int, but " "got: {}".format(
            sec.value, reason
        )


def test_validate_float_simple():
    bads = [
        PseudoSection("", "NotAFloat"),
    ]
    goods = [
        PseudoSection("", "0"),
        PseudoSection("", "1"),
        PseudoSection("", "-1"),
        PseudoSection("", "-0.1"),
        PseudoSection("", "0.1"),
        PseudoSection("", "100000000"),
        PseudoSection("", "-1000000000"),
        PseudoSection("", "+0"),
        PseudoSection("", "-0"),
    ]
    for sec in bads:
        valid, reason = con._validate_float(sec, sec.key)
        assert not valid, "{} should not have been a valid " "float".format(
            sec.value
        )
    for sec in goods:
        valid, reason = con._validate_float(sec, sec.key)
        assert (
            valid
        ), "{} should have been a valid float, but " "got: {}".format(
            sec.value, reason
        )


def test_validate_int_min():
    goods = [
        PseudoSection("", "0", mini=0),
        PseudoSection("", "1", mini=1),
        PseudoSection("", "-1", mini=-1),
    ]
    bads = [
        PseudoSection("", "1", mini=2),
        PseudoSection("", "0", mini=1),
    ]
    for sec in bads:
        valid, reason = con._validate_int(sec, sec.key, minimum=sec.mini)
        assert not valid, "{} should not have been a valid " "int".format(
            sec.value
        )
    for sec in goods:
        valid, reason = con._validate_int(sec, sec.key, minimum=sec.mini)
        assert valid, "{} should have been a valid int, but " "got: {}".format(
            sec.value, reason
        )


def test_validate_float_min():
    goods = [
        PseudoSection("", "0", mini=0.0),
        PseudoSection("", "0.1", mini=0.1),
        PseudoSection("", "-0.1", mini=-0.1),
        PseudoSection("", "0.1", mini=-0.1),
    ]
    bads = [
        PseudoSection("", "0.0999999999", mini=0.1),
    ]
    for sec in bads:
        valid, reason = con._validate_float(sec, sec.key, minimum=sec.mini)
        assert not valid, "{} should not have been a valid " "float".format(
            sec.value
        )
    for sec in goods:
        valid, reason = con._validate_float(sec, sec.key, minimum=sec.mini)
        assert (
            valid
        ), "{} should have been a valid float, but " "got: {}".format(
            sec.value, reason
        )


def test_validate_int_max():
    goods = [
        PseudoSection("", "0", maxi=0),
        PseudoSection("", "1", maxi=1),
        PseudoSection("", "-1", maxi=-1),
        PseudoSection("", "-1", maxi=1),
    ]
    bads = [
        PseudoSection("", "2", maxi=1),
        PseudoSection("", "1", maxi=0),
    ]
    for sec in bads:
        valid, reason = con._validate_int(sec, sec.key, maximum=sec.maxi)
        assert not valid, "{} should not have been a valid " "int".format(
            sec.value
        )
    for sec in goods:
        valid, reason = con._validate_int(sec, sec.key, maximum=sec.maxi)
        assert valid, "{} should have been a valid int, but " "got: {}".format(
            sec.value, reason
        )


def test_validate_float_max():
    goods = [
        PseudoSection("", "0", maxi=0.0),
        PseudoSection("", "0.1", maxi=0.1),
        PseudoSection("", "-0.1", maxi=-0.1),
        PseudoSection("", "-0.1", maxi=0.1),
    ]
    bads = [
        PseudoSection("", "0.10000000001", maxi=0.1),
    ]
    for sec in bads:
        valid, reason = con._validate_float(sec, sec.key, maximum=sec.maxi)
        assert not valid, "{} should not have been a valid " "float".format(
            sec.value
        )
    for sec in goods:
        valid, reason = con._validate_float(sec, sec.key, maximum=sec.maxi)
        assert (
            valid
        ), "{} should have been a valid float, but " "got: {}".format(
            sec.value, reason
        )


def test_validate_bool():
    goods = [
        "on",
        "True",
        "true",
        "yes",
        "off",
        "False",
        "false",
        "no",
        "0",
        "1",
    ]
    bads = [
        "onn",
        "offf",
        "2",
        "",
    ]
    for val in goods:
        conf = ConfigParser()
        conf.read_dict({"sec": {}})
        conf["sec"]["key"] = val
        valid, reason = con._validate_boolean(conf["sec"], "key")
        assert (
            valid
        ), "{} should have been a valid bool, but " "got: {}".format(
            val, reason
        )
    for val in bads:
        conf = ConfigParser()
        conf.read_dict({"sec": {}})
        conf["sec"]["key"] = val
        valid, reason = con._validate_boolean(conf["sec"], "key")
        assert not valid, "{} should not have been a valid " "bool".format(val)


def test_validate_url():
    goods = [
        "https://example.com",
        "https://example.com/",
        "https://example.com/foo.bar",
        "https://example.com/foo/bar",
        "https://user@example.com",
        "https://48.290.983.123:4443",
        "http://127.0.0.1:8000",
    ]
    bads = [
        "ftp://example.com/foo.bar",
        "http://",
        "http:///",
        "http://example.com",
    ]
    for val in goods:
        d = {"": val}
        valid, reason = con._validate_url(d, "")
        assert valid, "{} should have been a valid URL, but " "got: {}".format(
            val, reason
        )
    for val in bads:
        d = {"": val}
        valid, reason = con._validate_url(d, "")
        assert not valid, "{} should not have been a valid URL".format(val)


def test_nickname():
    max_len = 32
    goods = [
        "aaa",
        "AAA",
        "aAa",
        "A1a",
        "1aA",
        "aA1",
        "!!!",
        "!@#",
        "a!A",
        "!Aa",
        "Aa!",
        "a" * max_len,
    ]
    bads = [
        "",
        "a" * (max_len + 1),
        '"',
        "'",
    ]
    for nick in goods:
        d = {"n": nick}
        valid, reason = con._validate_nickname(d, "n")
        assert valid, reason
    for nick in bads:
        d = {"n": nick}
        valid, reason = con._validate_nickname(d, "n")
        assert not valid, reason


def test_country(conf):
    from string import Template

    err_tmpl = Template("$sec/$key ($val): $e")

    # Invalid default country code in scanner section
    errors = con._validate_country(conf, "scanner", "country", err_tmpl)
    assert (
        errors[0]
        == "scanner/country (AA): Not a valid ISO 3166 alpha-2 country code."
    )

    # Valid country code in scanner section
    conf["scanner"]["country"] = "US"
    errors = con._validate_country(conf, "scanner", "country", err_tmpl)
    assert not errors

    # No country in destinations.foo section
    conf["destinations"]["foo"] = "on"
    conf["destinations.foo"] = {}
    conf["destinations.foo"]["url"] = "https://foo.bar"
    errors = con._validate_country(
        conf, "destinations.foo", "country", err_tmpl
    )
    assert (
        errors[0] == "destinations.foo/country (None): "
        "Missing country in configuration file."
    )

    # Valid country in destinations.foo section
    conf["destinations.foo"]["url"] = "US"
    errors = con._validate_country(conf, "scanner", "country", err_tmpl)
    assert not errors


def test_config_arg_provided_but_no_found(args, conf):
    args.config = "non_existing_file"
    user_conf = con._get_user_config(args, conf)
    # since the user configuration is not found, it is the same as conf
    assert conf.__dict__.items() == user_conf.__dict__.items()


def test_config_arg_provided(args, conf, datadir):
    args.config = datadir.join("user_sbws.ini")
    user_conf = con._get_user_config(args, conf)
    assert user_conf["paths"]["sbws_home"] == "/tmp/.sbws"



"""Util functions to cleanup disk space."""
import gzip
import logging
import os
import shutil
import time
import types
from argparse import ArgumentDefaultsHelpFormatter
from datetime import datetime, timedelta

from sbws.globals import fail_hard
from sbws.util.filelock import DirectoryLock
from sbws.util.timestamp import unixts_to_dt_obj

log = logging.getLogger(__name__)


def gen_parser(sub):
    """
    Helper function for the broader argument parser generating code that adds
    in all the possible command line arguments for the cleanup command.

    :param argparse._SubParsersAction sub: what to add a sub-parser to
    """
    d = (
        "Compress and delete results and/or v3bw files old files."
        "Configuration options are read to determine which are old files"
    )
    p = sub.add_parser(
        "cleanup", description=d, formatter_class=ArgumentDefaultsHelpFormatter
    )
    p.add_argument(
        "--dry-run",
        action="store_true",
        help="Don't actually compress or delete anything",
    )
    p.add_argument(
        "--no-results", action="store_true", help="Do not clean results files"
    )
    p.add_argument(
        "--no-v3bw", action="store_true", help="Do not clean v3bw files"
    )


def _get_files_mtime_older_than(dname, days_delta, extensions):
    """Return files which modification time is older than days_delta
    and which extension is one of the extensions."""
    assert os.path.isdir(dname)
    assert isinstance(days_delta, int)
    assert isinstance(extensions, list)
    for ext in extensions:
        assert isinstance(ext, str)
        assert ext[0] == "."
    # Determine oldest allowed date
    today = datetime.utcfromtimestamp(time.time())
    oldest_day = today - timedelta(days=days_delta)
    for root, dirs, files in os.walk(dname):
        for f in files:
            fname = os.path.join(root, f)
            _, ext = os.path.splitext(fname)
            if ext not in extensions:
                log.debug(
                    "Ignoring %s because its extension is not in " "%s",
                    fname,
                    extensions,
                )
                continue
            # using file modification time instead of parsing the name
            # of the file.
            filedt = unixts_to_dt_obj(
                os.stat(fname, follow_symlinks=False).st_mtime
            )
            if filedt < oldest_day:
                yield fname


def _delete_files(dname, files, dry_run=True):
    """Delete the files passed as argument."""
    assert os.path.isdir(dname)
    assert isinstance(files, types.GeneratorType)
    with DirectoryLock(dname):
        for fname in files:
            log.info("Deleting %s", fname)
            assert os.path.commonprefix([dname, fname]) == dname
            if not dry_run:
                os.remove(fname)


def _compress_files(dname, files, dry_run=True):
    """Compress the files passed as argument."""
    assert os.path.isdir(dname)
    assert isinstance(files, types.GeneratorType)
    with DirectoryLock(dname):
        for fname in files:
            log.info("Compressing %s", fname)
            assert os.path.commonprefix([dname, fname]) == dname
            if dry_run:
                continue
            with open(fname, "rt") as in_fd:
                out_fname = fname + ".gz"
                with gzip.open(out_fname, "wt") as out_fd:
                    shutil.copyfileobj(in_fd, out_fd)
            os.remove(fname)


def _check_validity_periods_v3bw(compress_after_days, delete_after_days):
    if 1 <= compress_after_days and compress_after_days < delete_after_days:
        return True
    fail_hard(
        "v3bw files should only be compressed after 1 day and deleted "
        "after a bigger number of days."
    )


def _clean_v3bw_files(args, conf):
    v3bw_dname = conf.getpath("paths", "v3bw_dname")
    if not os.path.isdir(v3bw_dname):
        fail_hard("%s does not exist", v3bw_dname)
    compress_after_days = conf.getint(
        "cleanup", "v3bw_files_compress_after_days"
    )
    delete_after_days = conf.getint("cleanup", "v3bw_files_delete_after_days")
    _check_validity_periods_v3bw(compress_after_days, delete_after_days)
    # first delete so that the files to be deleted are not compressed first
    files_to_delete = _get_files_mtime_older_than(
        v3bw_dname, delete_after_days, [".v3bw", ".gz"]
    )
    _delete_files(v3bw_dname, files_to_delete, dry_run=args.dry_run)
    files_to_compress = _get_files_mtime_older_than(
        v3bw_dname, compress_after_days, [".v3bw"]
    )
    # when dry_run is true, compress will also show all the files that
    # would have been deleted, since they are not really deleted
    _compress_files(v3bw_dname, files_to_compress, dry_run=args.dry_run)


def _clean_result_files(args, conf):
    datadir = conf.getpath("paths", "datadir")
    if not os.path.isdir(datadir):
        fail_hard("%s does not exist", datadir)
    compress_after_days = conf.getint(
        "cleanup", "data_files_compress_after_days"
    )
    delete_after_days = conf.getint("cleanup", "data_files_delete_after_days")

    # first delete so that the files to be deleted are not compressed first
    files_to_delete = _get_files_mtime_older_than(
        datadir, delete_after_days, [".txt", ".gz"]
    )
    _delete_files(datadir, files_to_delete, dry_run=args.dry_run)

    # when dry_run is true, compress will also show all the files that
    # would have been deleted, since they are not really deleted
    files_to_compress = _get_files_mtime_older_than(
        datadir, compress_after_days, [".txt"]
    )
    _compress_files(datadir, files_to_compress, dry_run=args.dry_run)


def main(args, conf):
    """
    Main entry point in to the cleanup command.

    :param argparse.Namespace args: command line arguments
    :param configparser.ConfigParser conf: parsed config files
    """
    if not args.no_results:
        _clean_result_files(args, conf)

    if not args.no_v3bw:
        _clean_v3bw_files(args, conf)



# -*- coding: utf-8 -*-
"""Classes and functions that create the bandwidth measurements document
(v3bw) used by bandwidth authorities."""
# flake8: noqa: E741
# (E741 ambiguous variable name), when using l.

import copy
import logging
import math
import os
from itertools import combinations
from statistics import mean, median

from stem.descriptor import parse_file

from sbws import __version__
from sbws.globals import (
    BW_LINE_SIZE,
    MAX_BW_DIFF_PERC,
    MIN_REPORT,
    PROP276_ROUND_DIG,
    SBWS_SCALE_CONSTANT,
    SBWS_SCALING,
    SPEC_VERSION,
    TORFLOW_BW_MARGIN,
    TORFLOW_OBS_LAST,
    TORFLOW_OBS_MEAN,
    TORFLOW_SCALING,
)
from sbws.lib import scaling
from sbws.lib.resultdump import ResultSuccess, _ResultType
from sbws.util.filelock import DirectoryLock
from sbws.util.state import State
from sbws.util.timestamp import (
    dt_obj_to_isodt_str,
    isostr_to_dt_obj,
    now_isodt_str,
    now_unixts,
    unixts_to_isodt_str,
)

log = logging.getLogger(__name__)

LINE_SEP = "\n"
KEYVALUE_SEP_V1 = "="
KEYVALUE_SEP_V2 = " "

# NOTE: in a future refactor make make all the KeyValues be a dictionary
# with their type, so that it's more similar to stem parser.

# Header KeyValues
# =================
# KeyValues that need to be in a specific order in the Bandwidth File.
HEADER_KEYS_V1_1_ORDERED = ["version"]
# KeyValues that are not initialized from the state file nor the measurements.
# They can also be pass as an argument to `Header` to overwrite default values,
# what is done in unit tests.
# `latest bandwidth` is special cause it gets its value from timestamp, which
# is not a KeyValue, but it's always pass as an argument.
# It could be separaed in other list, but so far there is no need, cause:
# 1. when it's pass to the Header to initialize it, it's just ignored.
# 2. when the file is created, it's took into account.
HEADER_KEYS_V1_1_SELF_INITIALIZED = [
    "software",
    "software_version",
    "file_created",
    "latest_bandwidth",
]
# KeyValues that are initialized from arguments.
HEADER_KEYS_V1_1_TO_INIT = [
    "earliest_bandwidth",
    "generator_started",
]

# number_eligible_relays is the number that ends in the bandwidth file
# ie, have not been excluded by one of the filters in 4. below
# They should be call recent_measurement_included_count to be congruent
# with the other KeyValues.
HEADER_KEYS_V1_2 = [
    "number_eligible_relays",
    "minimum_number_eligible_relays",
    "number_consensus_relays",
    "percent_eligible_relays",
    "minimum_percent_eligible_relays",
]

# KeyValues added in the Bandwidth File v1.3.0
HEADER_KEYS_V1_3 = [
    "scanner_country",
    "destinations_countries",
]

# KeyValues that count the number of relays that are in the bandwidth file,
# but ignored by Tor when voting, because they do not have a
# measured bandwidth.
HEADER_RECENT_MEASUREMENTS_EXCLUDED_KEYS = [
    # Number of relays that were measured but all the measurements failed
    # because of network failures or it was
    # not found a suitable helper relay
    "recent_measurements_excluded_error_count",
    # Number of relays that have successful measurements but the measurements
    # were not away from each other in X time (by default 1 day).
    "recent_measurements_excluded_near_count",
    # Number of relays that have successful measurements and they are away from
    # each other but they are not X time recent.
    # By default this is 5 days, which is the same time the older
    # the measurements can be by default.
    "recent_measurements_excluded_old_count",
    # Number of relays that have successful measurements and they are away from
    # each other and recent
    # but the number of measurements are less than X (by default 2).
    "recent_measurements_excluded_few_count",
]
# Added in #29591
# NOTE: recent_consensus_count, recent_priority_list_count,
# recent_measurement_attempt_count and recent_priority_relay_count
# are not reset when the scanner is stop.
# They will accumulate the values since the scanner was ever started.
HEADER_KEYS_V1_4 = [
    # 1.1 header: the number of different consensuses, that sbws has seen,
    # since the last 5 days
    "recent_consensus_count",
    # 2.4 Number of times a priority list has been created
    "recent_priority_list_count",
    # 2.5 Number of relays that there were in a priority list
    # [50, number of relays in the network * 0.05]
    "recent_priority_relay_count",
    # 3.6 header: the number of times that sbws has tried to measure any relay,
    # since the last 5 days
    # This would be the number of times a relays were in a priority list
    "recent_measurement_attempt_count",
    # 3.7 header: the number of times that sbws has tried to measure any relay,
    # since the last 5 days, but it didn't work
    # This should be the number of attempts - number of ResultSuccess -
    # something else we don't know yet
    # So far is the number of ResultError
    "recent_measurement_failure_count",
    # The time it took to report about half of the network.
    "time_to_report_half_network",
] + HEADER_RECENT_MEASUREMENTS_EXCLUDED_KEYS

# Tor version will be obtained from the state file, so it won't be pass as an
# argument, but will be self-initialized.
HEADER_KEYS_V1_4_TO_INIT = ["tor_version"]

# KeyValues that are initialized from arguments, not self-initialized.
HEADER_INIT_KEYS = (
    HEADER_KEYS_V1_1_TO_INIT
    + HEADER_KEYS_V1_3
    + HEADER_KEYS_V1_2
    + HEADER_KEYS_V1_4
    + HEADER_KEYS_V1_4_TO_INIT
)

HEADER_INT_KEYS = HEADER_KEYS_V1_2 + HEADER_KEYS_V1_4
# List of all unordered KeyValues currently being used to generate the file
HEADER_UNORDERED_KEYS = (
    HEADER_KEYS_V1_1_SELF_INITIALIZED
    + HEADER_KEYS_V1_1_TO_INIT
    + HEADER_KEYS_V1_3
    + HEADER_KEYS_V1_2
    + HEADER_KEYS_V1_4
    + HEADER_KEYS_V1_4_TO_INIT
)
# List of all the KeyValues currently being used to generate the file
HEADER_ALL_KEYS = HEADER_KEYS_V1_1_ORDERED + HEADER_UNORDERED_KEYS

TERMINATOR = "====="

# Bandwidth Lines KeyValues
# =========================
# Num header lines in v1.X.X using all the KeyValues
NUM_LINES_HEADER_V1 = len(HEADER_ALL_KEYS) + 2
LINE_TERMINATOR = TERMINATOR + LINE_SEP

# KeyValue separator in Bandwidth Lines
BWLINE_KEYVALUES_SEP_V1 = " "
# not including in the files the extra bws for now
BWLINE_KEYS_V0 = ["node_id", "bw"]
BWLINE_KEYS_V1_1 = [
    "master_key_ed25519",
    "nick",
    "rtt",
    "time",
    "success",
    "error_stream",
    "error_circ",
    "error_misc",
    # Added in #292951
    "error_second_relay",
    "error_destination",
]
BWLINE_KEYS_V1_2 = [
    "bw_median",
    "bw_mean",
    "desc_bw_avg",
    "desc_bw_bur",
    "desc_bw_obs_last",
    "desc_bw_obs_mean",
    "consensus_bandwidth",
    "consensus_bandwidth_is_unmeasured",
]

# There were no bandwidth lines key added in the specification version 1.3

# Added in #292951
BWLINE_KEYS_V1_4 = [
    # 1.2 relay: the number of different consensuses, that sbws has seen,
    # since the last 5 days, that have this relay
    "relay_in_recent_consensus_count",
    # 2.6 relay: the number of times a relay was "prioritized" to be measured
    # in the recent days (by default 5).
    "relay_recent_priority_list_count",
    # 3.8 relay:  the number of times that sbws has tried to measure
    # this relay, since the last 5 days
    # This would be the number of times a relay was in a priority list (2.6)
    # since once it gets measured, it either returns ResultError,
    # ResultSuccess or something else happened that we don't know yet
    "relay_recent_measurement_attempt_count",
    # 3.9 relay:  the number of times that sbws has tried to measure
    # this relay, since the last 5 days, but it didn't work
    # This should be the number of attempts - number of ResultSuccess -
    # something else we don't know yet
    # So far is the number of ResultError
    "relay_recent_measurement_failure_count",
    # Number of error results created in the last 5 days that are excluded.
    # This is the sum of all the errors.
    "relay_recent_measurements_excluded_error_count",
    # The number of successful results, created in the last 5 days,
    # that were excluded by a rule, for this relay.
    # 'relay_recent_measurements_excluded_error_count' would be the
    # sum of the following 3 + the number of error results.
    # The number of successful measurements that are not X time away
    # from each other (by default 1 day).
    "relay_recent_measurements_excluded_near_count",
    # The number of successful measurements that are away from each other
    # but not X time recent (by default 5 days).
    "relay_recent_measurements_excluded_old_count",
    # The number of measurements excluded because they are not at least X
    # (by default 2).
    "relay_recent_measurements_excluded_few_count",
    # `vote=0` is used for the relays that were excluded to
    # be reported in the bandwidth file and now they are
    # reported.
    # It tells Tor to do not vote on the relay.
    # `unmeasured=1` is used for the same relays and it is
    # added in case Tor would vote on them in future versions.
    # Maybe these keys should not be included for the relays
    # in which vote=1 and unmeasured=0.
    "vote",
    "unmeasured",
    # When there not enough eligible relays (not excluded)
    # under_min_report is 1, `vote` is 0.
    # Added in #29853.
    "under_min_report",
]
BWLINE_KEYS_V1 = (
    BWLINE_KEYS_V0 + BWLINE_KEYS_V1_1 + BWLINE_KEYS_V1_2 + BWLINE_KEYS_V1_4
)
# NOTE: tech-debt: assign boolean type to vote and unmeasured,
# when the attributes are defined with a type, as stem does.
BWLINE_INT_KEYS = (
    [
        "bw",
        "rtt",
        "success",
        "error_stream",
        "error_circ",
        "error_misc",
    ]
    + BWLINE_KEYS_V1_2
    + BWLINE_KEYS_V1_4
)
# This is boolean, not int.
BWLINE_INT_KEYS.remove("consensus_bandwidth_is_unmeasured")


def round_sig_dig(n, digits=PROP276_ROUND_DIG):
    """Round n to 'digits' significant digits in front of the decimal point.
    Results less than or equal to 1 are rounded to 1.
    Returns an integer.

    digits must be greater than 0.
    n must be less than or equal to 2**73, to avoid floating point errors.
    """
    digits = int(digits)
    assert digits >= 1
    if n <= 1:
        return 1
    digits_in_n = int(math.log10(n)) + 1
    round_digits = max(digits_in_n - digits, 0)
    rounded_n = round(n, -round_digits)
    return int(rounded_n)


def kb_round_x_sig_dig(bw_bs, digits=PROP276_ROUND_DIG):
    """Convert bw_bs from bytes to kilobytes, and round the result to
    'digits' significant digits.
    Results less than or equal to 1 are rounded up to 1.
    Returns an integer.

    digits must be greater than 0.
    n must be less than or equal to 2**82, to avoid floating point errors.
    """
    # avoid double-rounding by using floating-point
    bw_kb = bw_bs / 1000.0
    return round_sig_dig(bw_kb, digits=digits)


def num_results_of_type(results, type_str):
    return len([r for r in results if r.type == type_str])


# Better way to use enums?
def result_type_to_key(type_str):
    return type_str.replace("-", "_")


class V3BWHeader(object):
    """
    Create a bandwidth measurements (V3bw) header
    following bandwidth measurements document spec version 1.X.X.

    :param str timestamp: timestamp in Unix Epoch seconds of the most recent
        generator result.
    :param str version: the spec version
    :param str software: the name of the software that generates this
    :param str software_version: the version of the software
    :param dict kwargs: extra headers. Currently supported:

        - earliest_bandwidth: str, ISO 8601 timestamp in UTC time zone
          when the first bandwidth was obtained
        - generator_started: str, ISO 8601 timestamp in UTC time zone
          when the generator started
    """

    def __init__(self, timestamp, **kwargs):
        assert isinstance(timestamp, str)
        for v in kwargs.values():
            assert isinstance(v, str)
        self.timestamp = timestamp
        # KeyValues with default value when not given by kwargs
        self.version = kwargs.get("version", SPEC_VERSION)
        self.software = kwargs.get("software", "sbws")
        self.software_version = kwargs.get("software_version", __version__)
        self.file_created = kwargs.get("file_created", now_isodt_str())
        # latest_bandwidth should not be in kwargs, since it MUST be the
        # same as timestamp
        self.latest_bandwidth = unixts_to_isodt_str(timestamp)
        [
            setattr(self, k, v)
            for k, v in kwargs.items()
            if k in HEADER_INIT_KEYS
        ]

    def __str__(self):
        if self.version.startswith("1."):
            return self.strv1
        return self.strv2

    @classmethod
    def from_results(
        cls,
        results,
        scanner_country=None,
        destinations_countries=None,
        state_fpath="",
    ):
        kwargs = dict()
        latest_bandwidth = cls.latest_bandwidth_from_results(results)
        earliest_bandwidth = cls.earliest_bandwidth_from_results(results)
        # NOTE: Blocking, reads file
        generator_started = cls.generator_started_from_file(state_fpath)
        recent_consensus_count = cls.consensus_count_from_file(state_fpath)
        timestamp = str(latest_bandwidth)

        # XXX: tech-debt: obtain the other values from the state file using
        # this state variable.
        # Store the state as an attribute of the object?
        state = State(state_fpath)
        tor_version = state.get("tor_version", None)
        if tor_version:
            kwargs["tor_version"] = tor_version

        kwargs["latest_bandwidth"] = unixts_to_isodt_str(latest_bandwidth)
        kwargs["earliest_bandwidth"] = unixts_to_isodt_str(earliest_bandwidth)
        if generator_started is not None:
            kwargs["generator_started"] = generator_started
        # To be compatible with older bandwidth files, do not require it.
        if scanner_country is not None:
            kwargs["scanner_country"] = scanner_country
        if destinations_countries is not None:
            kwargs["destinations_countries"] = destinations_countries
        if recent_consensus_count is not None:
            kwargs["recent_consensus_count"] = recent_consensus_count

        recent_measurement_attempt_count = (
            cls.recent_measurement_attempt_count_from_file(state_fpath)
        )
        if recent_measurement_attempt_count is not None:
            kwargs["recent_measurement_attempt_count"] = str(
                recent_measurement_attempt_count
            )

        # If it is a failure that is not a ResultError, then
        # failures = attempts - all mesaurements
        # Works only in the case that old measurements files already had
        # measurements count
        # If this is None or 0, the failures can't be calculated
        if recent_measurement_attempt_count:
            all_measurements = 0
            for result_list in results.values():
                all_measurements += len(result_list)
            measurement_failures = (
                recent_measurement_attempt_count - all_measurements
            )
            kwargs["recent_measurement_failure_count"] = str(
                measurement_failures
            )

        priority_lists = cls.recent_priority_list_count_from_file(state_fpath)
        if priority_lists is not None:
            kwargs["recent_priority_list_count"] = str(priority_lists)

        priority_relays = cls.recent_priority_relay_count_from_file(
            state_fpath
        )
        if priority_relays is not None:
            kwargs["recent_priority_relay_count"] = str(priority_relays)

        h = cls(timestamp, **kwargs)
        return h

    @classmethod
    def from_lines_v1(cls, lines):
        """
        :param list lines: list of lines to parse
        :returns: tuple of V3BWHeader object and non-header lines
        """
        assert isinstance(lines, list)
        try:
            index_terminator = lines.index(TERMINATOR)
        except ValueError:
            # is not a bw file or is v100
            log.warn("Terminator is not in lines")
            return None
        ts = lines[0]
        kwargs = dict(
            [
                l.split(KEYVALUE_SEP_V1)
                for l in lines[:index_terminator]
                if l.split(KEYVALUE_SEP_V1)[0] in HEADER_ALL_KEYS
            ]
        )
        h = cls(ts, **kwargs)
        # last line is new line
        return h, lines[index_terminator + 1 : -1]

    @classmethod
    def from_text_v1(self, text):
        """
        :param str text: text to parse
        :returns: tuple of V3BWHeader object and non-header lines
        """
        assert isinstance(text, str)
        return self.from_lines_v1(text.split(LINE_SEP))

    @classmethod
    def from_lines_v100(cls, lines):
        """
        :param list lines: list of lines to parse
        :returns: tuple of V3BWHeader object and non-header lines
        """
        assert isinstance(lines, list)
        h = cls(lines[0])
        # last line is new line
        return h, lines[1:-1]

    @staticmethod
    def generator_started_from_file(state_fpath):
        """
        ISO formatted timestamp for the time when the scanner process most
        recently started.
        """
        state = State(state_fpath)
        if "scanner_started" in state:
            # From v1.1.0-dev `state` is capable of converting strs to datetime
            return dt_obj_to_isodt_str(state["scanner_started"])
        else:
            return None

    @staticmethod
    def consensus_count_from_file(state_fpath):
        state = State(state_fpath)
        count = state.count("recent_consensus")
        if count:
            return str(count)
        return None

    # NOTE: in future refactor store state in the class
    @staticmethod
    def recent_measurement_attempt_count_from_file(state_fpath):
        """
        Returns the number of times any relay was queued to be measured
        in the recent (by default 5) days from the state file.
        """
        state = State(state_fpath)
        return state.count("recent_measurement_attempt")

    @staticmethod
    def recent_priority_list_count_from_file(state_fpath):
        """
        Returns the number of times
        :meth:`~sbws.lib.relayprioritizer.RelayPrioritizer.best_priority`
        was run
        in the recent (by default 5) days from the state file.
        """
        state = State(state_fpath)
        return state.count("recent_priority_list")

    @staticmethod
    def recent_priority_relay_count_from_file(state_fpath):
        """
        Returns the number of times any relay was "prioritized" to be measured
        in the recent (by default 5) days from the state file.
        """
        state = State(state_fpath)
        return state.count("recent_priority_relay")

    @staticmethod
    def latest_bandwidth_from_results(results):
        return round(max([r.time for fp in results for r in results[fp]]))

    @staticmethod
    def earliest_bandwidth_from_results(results):
        return round(min([r.time for fp in results for r in results[fp]]))

    @property
    def keyvalue_unordered_tuple_ls(self):
        """Return list of KeyValue tuples that do not have specific order."""
        # sort the list to generate determinist headers
        keyvalue_tuple_ls = sorted(
            [
                (k, v)
                for k, v in self.__dict__.items()
                if k in HEADER_UNORDERED_KEYS
            ]
        )
        return keyvalue_tuple_ls

    @property
    def keyvalue_tuple_ls(self):
        """Return list of all KeyValue tuples"""
        return [("version", self.version)] + self.keyvalue_unordered_tuple_ls

    @property
    def keyvalue_v1str_ls(self):
        """Return KeyValue list of strings following spec v1.X.X."""
        keyvalues = [self.timestamp] + [
            KEYVALUE_SEP_V1.join([k, v]) for k, v in self.keyvalue_tuple_ls
        ]
        return keyvalues

    @property
    def strv1(self):
        """Return header string following spec v1.X.X."""
        header_str = (
            LINE_SEP.join(self.keyvalue_v1str_ls) + LINE_SEP + LINE_TERMINATOR
        )
        return header_str

    @property
    def keyvalue_v2_ls(self):
        """Return KeyValue list of strings following spec v2.X.X."""
        keyvalue = [self.timestamp] + [
            KEYVALUE_SEP_V2.join([k, v]) for k, v in self.keyvalue_tuple_ls
        ]
        return keyvalue

    @property
    def strv2(self):
        """Return header string following spec v2.X.X."""
        header_str = (
            LINE_SEP.join(self.keyvalue_v2_ls) + LINE_SEP + LINE_TERMINATOR
        )
        return header_str

    @property
    def num_lines(self):
        return len(self.__str__().split(LINE_SEP))

    def add_stats(self, **kwargs):
        # Using kwargs because attributes might change.
        [
            setattr(self, k, str(v))
            for k, v in kwargs.items()
            if k in HEADER_KEYS_V1_2
        ]

    def add_time_report_half_network(self):
        """Add to the header the time it took to measure half of the network.

        It is not the time the scanner actually takes on measuring all the
        network, but the ``number_eligible_relays`` that are reported in the
        bandwidth file and directory authorities will vote on.

        This is calculated for half of the network, so that failed or not
        reported relays do not affect too much.

        For instance, if there are 6500 relays in the network, half of the
        network would be 3250. And if there were 4000 eligible relays
        measured in an interval of 3 days, the time to measure half of the
        network would be 3 days * 3250 / 4000.

        Since the elapsed time is calculated from the earliest and the
        latest measurement and a relay might have more than 2 measurements,
        this would give an estimate on how long it would take to measure
        the network including all the valid measurements.

        Log also an estimated on how long it would take with the current
        number of relays included in the bandwidth file.
        """
        # NOTE: in future refactor do not convert attributes to str until
        # writing to the file, so that they do not need to be converted back
        # to do some calculations.
        elapsed_time = (
            isostr_to_dt_obj(self.latest_bandwidth)
            - isostr_to_dt_obj(self.earliest_bandwidth)
        ).total_seconds()

        # This attributes were added later and some tests that
        # do not initialize them would fail.
        eligible_relays = int(getattr(self, "number_eligible_relays", 0))
        consensus_relays = int(getattr(self, "number_consensus_relays", 0))
        if not (eligible_relays and consensus_relays):
            return

        half_network = consensus_relays / 2
        # Calculate the time it would take to measure half of the network
        if eligible_relays >= half_network:
            time_half_network = round(
                elapsed_time * half_network / eligible_relays
            )
            self.time_to_report_half_network = str(time_half_network)

        # In any case log an estimated on the time to measure all the network.
        estimated_time = round(
            elapsed_time * consensus_relays / eligible_relays
        )
        log.info(
            "Estimated time to successfully report about all the relays in the"
            " last consensus: %s hours.",
            round(estimated_time / 60 / 60),
        )

    def add_relays_excluded_counters(self, exclusion_dict):
        """
        Add the monitoring KeyValues to the header about the number of
        relays not included because they were not ``eligible``.
        """
        log.debug("Adding relays excluded counters.")
        for k, v in exclusion_dict.items():
            setattr(self, k, str(v))


class V3BWLine(object):
    """
    Create a Bandwidth List line following the spec version 1.X.X.

    :param str node_id: the relay fingerprint
    :param int bw: the bandwidth value that directory authorities will include
        in their votes.
    :param dict kwargs: extra headers.

    .. note:: tech-debt: move node_id and bw to kwargs and just ensure that
       the required values are in ``**kwargs``
    """

    def __init__(self, node_id, bw, **kwargs):
        assert isinstance(node_id, str)
        assert node_id.startswith("$")
        self.node_id = node_id
        self.bw = bw
        # For now, we do not want to add ``bw_filt`` to the bandwidth file,
        # therefore it is set here but not added to ``BWLINE_KEYS_V1``.
        [
            setattr(self, k, v)
            for k, v in kwargs.items()
            if k in BWLINE_KEYS_V1 + ["bw_filt"]
        ]

    def __str__(self):
        return self.bw_strv1

    @classmethod
    def from_results(
        cls,
        results,
        secs_recent=None,
        secs_away=None,
        min_num=0,
        router_statuses_d=None,
    ):
        """Convert sbws results to relays' Bandwidth Lines

        ``bs`` stands for Bytes/seconds
        ``bw_mean`` means the bw is obtained from the mean of the all the
        downloads' bandwidth.
        Downloads' bandwidth are calculated as the amount of data received
        divided by the the time it took to received.
        bw = data (Bytes) / time (seconds)
        """
        # log.debug("Len success_results %s", len(success_results))
        node_id = "$" + results[0].fingerprint
        kwargs = dict()
        kwargs["nick"] = results[0].nickname
        if getattr(results[0], "master_key_ed25519"):
            kwargs["master_key_ed25519"] = results[0].master_key_ed25519
        kwargs["time"] = cls.last_time_from_results(results)
        kwargs.update(cls.result_types_from_results(results))

        # If it has not the attribute, return list to be able to call len
        # If it has the attribute, but it is None, return also list
        kwargs["relay_in_recent_consensus_count"] = str(
            max(
                [
                    len(getattr(r, "relay_in_recent_consensus", []) or [])
                    for r in results
                ]
            )
        )

        # Workaround for #34309.
        # Because of a bug, probably in relaylist, resultdump, relayprioritizer
        # or scanner, only the last timestamp is being stored in each result.
        # Temporally count the number of timestamps for all results.
        # If there is an unexpected failure and the result is not stored, this
        # number would be lower than what would be the correct one.
        # This should happen rarely or never.
        ts = set([])
        for r in results:
            if getattr(r, "relay_recent_priority_list", None):
                ts.update(r.relay_recent_priority_list)
        kwargs["relay_recent_priority_list_count"] = str(len(ts))

        # Same comment as the previous paragraph.
        ts = set()
        for r in results:
            if getattr(r, "relay_recent_measurement_attempt", None):
                ts.update(r.relay_recent_measurement_attempt)
        kwargs["relay_recent_measurement_attempt_count"] = str(len(ts))

        success_results = [r for r in results if isinstance(r, ResultSuccess)]

        # NOTE: The following 4 conditions exclude relays from the bandwidth
        # file when the measurements does not satisfy some rules, what makes
        # the relay non-`eligible`.
        # In BWLINE_KEYS_V1_4 it is explained what they mean.
        # In HEADER_RECENT_MEASUREMENTS_EXCLUDED_KEYS it is also
        # explained the what it means the strings returned.
        # They rules were introduced in #28061 and #27338
        # In #28565 we introduce the KeyValues to know why they're excluded.
        # In #28563 we report these relays, but make Tor ignore them.
        # This might confirm #28042.

        # If the relay is non-`eligible`:
        # Create a bandwidth line with the relay, but set ``vote=0`` so that
        # Tor versions with patch #29806 does not vote on the relay.
        # Set ``bw=1`` so that Tor versions without the patch,
        # will give the relay low bandwidth.
        # Include ``unmeasured=1`` in case Tor would vote on unmeasured relays
        # in future versions.
        # And return because there are not bandwidth values.
        # NOTE: the bandwidth values could still be obtained if:
        # 1. ``ResultError`` will store them
        # 2. assign ``results_recent = results`` when there is a ``exclusion
        # reason.
        # This could be done in a better way as part of a refactor #28684.

        kwargs["vote"] = 0
        kwargs["unmeasured"] = 1

        exclusion_reason = None

        number_excluded_error = len(results) - len(success_results)
        if number_excluded_error > 0:
            # then the number of error results is the number of results
            kwargs[
                "relay_recent_measurements_excluded_error_count"
            ] = number_excluded_error
        if not success_results:
            exclusion_reason = "recent_measurements_excluded_error_count"
            return (cls(node_id, 1, **kwargs), exclusion_reason)

        results_away = cls.results_away_each_other(success_results, secs_away)
        number_excluded_near = len(success_results) - len(results_away)
        if number_excluded_near > 0:
            kwargs[
                "relay_recent_measurements_excluded_near_count"
            ] = number_excluded_near
        if not results_away:
            exclusion_reason = "recent_measurements_excluded_near_count"
            return (cls(node_id, 1, **kwargs), exclusion_reason)
        # log.debug("Results away from each other: %s",
        #           [unixts_to_isodt_str(r.time) for r in results_away])

        results_recent = cls.results_recent_than(results_away, secs_recent)
        number_excluded_old = len(results_away) - len(results_recent)
        if number_excluded_old > 0:
            kwargs[
                "relay_recent_measurements_excluded_old_count"
            ] = number_excluded_old
        if not results_recent:
            exclusion_reason = "recent_measurements_excluded_old_count"
            return (cls(node_id, 1, **kwargs), exclusion_reason)

        if not len(results_recent) >= min_num:
            kwargs["relay_recent_measurements_excluded_few_count"] = len(
                results_recent
            )
            # log.debug('The number of results is less than %s', min_num)
            exclusion_reason = "recent_measurements_excluded_few_count"
            return (cls(node_id, 1, **kwargs), exclusion_reason)

        # Use the last consensus if available, since the results' consensus
        # values come from the moment the measurement was made.
        if router_statuses_d and node_id in router_statuses_d:
            consensus_bandwidth = router_statuses_d[node_id].bandwidth * 1000
            consensus_bandwidth_is_unmeasured = router_statuses_d[
                node_id
            ].is_unmeasured
        else:
            consensus_bandwidth = cls.consensus_bandwidth_from_results(
                results_recent
            )
            consensus_bandwidth_is_unmeasured = (
                cls.consensus_bandwidth_is_unmeasured_from_results(
                    results_recent
                )
            )
        # If there is no last observed bandwidth, there won't be mean either.
        desc_bw_obs_last = cls.desc_bw_obs_last_from_results(results_recent)

        # Exclude also relays without observed bandwidth, since they can't be
        # scaled
        if desc_bw_obs_last is None:
            # This reason is not counted, not added in the file, but it will
            # have vote = 0
            return (cls(node_id, 1), "no_observed_bw")

        # For any line not excluded, do not include vote and unmeasured
        # KeyValues
        del kwargs["vote"]
        del kwargs["unmeasured"]

        rtt = cls.rtt_from_results(results_recent)
        if rtt:
            kwargs["rtt"] = rtt
        bw = cls.bw_median_from_results(results_recent)
        # XXX: all the class functions could use the bw_measurements instead of
        # obtaining them each time or use a class Measurements.
        bw_measurements = scaling.bw_measurements_from_results(results_recent)
        kwargs["bw_mean"] = cls.bw_mean_from_results(results_recent)
        kwargs["bw_filt"] = scaling.bw_filt(bw_measurements)
        kwargs["bw_median"] = cls.bw_median_from_results(results_recent)
        kwargs["desc_bw_avg"] = cls.desc_bw_avg_from_results(results_recent)
        kwargs["desc_bw_bur"] = cls.desc_bw_bur_from_results(results_recent)
        kwargs["consensus_bandwidth"] = consensus_bandwidth
        kwargs[
            "consensus_bandwidth_is_unmeasured"
        ] = consensus_bandwidth_is_unmeasured
        kwargs["desc_bw_obs_last"] = desc_bw_obs_last
        kwargs["desc_bw_obs_mean"] = cls.desc_bw_obs_mean_from_results(
            results_recent
        )

        bwl = cls(node_id, bw, **kwargs)
        return bwl, None

    @classmethod
    def from_data(cls, data, fingerprint):
        assert fingerprint in data
        return cls.from_results(data[fingerprint])

    @classmethod
    def from_bw_line_v1(cls, line):
        assert isinstance(line, str)
        kwargs = dict(
            [
                kv.split(KEYVALUE_SEP_V1)
                for kv in line.split(BWLINE_KEYVALUES_SEP_V1)
                if kv.split(KEYVALUE_SEP_V1)[0] in BWLINE_KEYS_V1
            ]
        )
        for k, v in kwargs.items():
            if k in BWLINE_INT_KEYS:
                kwargs[k] = int(v)
        node_id = kwargs["node_id"]
        bw = kwargs["bw"]
        del kwargs["node_id"]
        del kwargs["bw"]
        bw_line = cls(node_id, bw, **kwargs)
        return bw_line

    @staticmethod
    def results_away_each_other(results, secs_away=None):
        # log.debug("Checking whether results are away from each other in %s "
        #           "secs.", secs_away)
        if secs_away is None or len(results) < 2:
            return results
        for a, b in combinations(results, 2):
            if abs(a.time - b.time) > secs_away:
                return results
        # log.debug("Results are NOT away from each other in at least %ss: %s",
        #           secs_away, [unixts_to_isodt_str(r.time) for r in results])
        return []

    @staticmethod
    def results_recent_than(results, secs_recent=None):
        if secs_recent is None:
            return results
        results_recent = list(
            filter(lambda x: (now_unixts() - x.time) < secs_recent, results)
        )
        # if not results_recent:
        #     log.debug("Results are NOT more recent than %ss: %s",
        #               secs_recent,
        #               [unixts_to_isodt_str(r.time) for r in results])
        return results_recent

    @staticmethod
    def bw_median_from_results(results):
        bws = [
            dl["amount"] / dl["duration"]
            for r in results
            for dl in r.downloads
        ]
        if bws:
            return max(round(median(bws)), 1)
        return 1

    @staticmethod
    def bw_mean_from_results(results):
        bws = [
            dl["amount"] / dl["duration"]
            for r in results
            for dl in r.downloads
        ]
        # It's safe to return 0 here, because:
        # 1. this value will be the numerator when calculating the ratio.
        # 2. `kb_round_x_sig_dig` returns a minimum of 1.
        if bws:
            return round(mean(bws))
        return 0

    @staticmethod
    def last_time_from_results(results):
        return unixts_to_isodt_str(round(max([r.time for r in results])))

    @staticmethod
    def rtt_from_results(results):
        # convert from milliseconds to seconds
        rtts = [(round(rtt * 1000)) for r in results for rtt in r.rtts]
        rtt = round(median(rtts)) if rtts else None
        return rtt

    @staticmethod
    def result_types_from_results(results):
        rt_dict = dict(
            [
                (
                    result_type_to_key(rt.value),
                    num_results_of_type(results, rt.value),
                )
                for rt in _ResultType
            ]
        )
        return rt_dict

    @staticmethod
    def desc_bw_avg_from_results(results):
        """Obtain the last descriptor bandwidth average from the results."""
        for r in reversed(results):
            if r.relay_average_bandwidth is not None:
                return r.relay_average_bandwidth
        log.warning("Descriptor average bandwidth is None.")
        return None

    @staticmethod
    def desc_bw_bur_from_results(results):
        """Obtain the last descriptor bandwidth burst from the results."""
        for r in reversed(results):
            if r.relay_burst_bandwidth is not None:
                return r.relay_burst_bandwidth
        log.warning("Descriptor burst bandwidth is None.")
        return None

    @staticmethod
    def consensus_bandwidth_from_results(results):
        """Obtain the last consensus bandwidth from the results."""
        for r in reversed(results):
            if r.consensus_bandwidth is not None:
                return r.consensus_bandwidth
        log.warning("Consensus bandwidth is None.")
        return None

    @staticmethod
    def consensus_bandwidth_is_unmeasured_from_results(results):
        """Obtain the last consensus unmeasured flag from the results."""
        for r in reversed(results):
            if r.consensus_bandwidth_is_unmeasured is not None:
                return r.consensus_bandwidth_is_unmeasured
            log.warning("Consensus bandwidth is unmeasured is None.")
        return None

    @staticmethod
    def desc_bw_obs_mean_from_results(results):
        desc_bw_obs_ls = []
        for r in results:
            if r.relay_observed_bandwidth is not None:
                desc_bw_obs_ls.append(r.relay_observed_bandwidth)
        if desc_bw_obs_ls:
            return round(mean(desc_bw_obs_ls))
        log.info(
            "Descriptor observed bandwidth is None. Can't calculate mean."
        )
        return None

    @staticmethod
    def desc_bw_obs_last_from_results(results):
        # the last is at the end of the list
        for r in reversed(results):
            if r.relay_observed_bandwidth is not None:
                return r.relay_observed_bandwidth
        log.info("Descriptor observed bandwidth is None. Can't get the last.")
        return None

    @property
    def bw_keyvalue_tuple_ls(self):
        """Return list of KeyValue Bandwidth Line tuples."""
        # sort the list to generate determinist headers
        keyvalue_tuple_ls = sorted(
            [(k, v) for k, v in self.__dict__.items() if k in BWLINE_KEYS_V1]
        )
        return keyvalue_tuple_ls

    @property
    def bw_keyvalue_v1str_ls(self):
        """Return list of KeyValue Bandwidth Line strings following
        spec v1.X.X.
        """
        bw_keyvalue_str = [
            KEYVALUE_SEP_V1.join([k, str(v)])
            for k, v in self.bw_keyvalue_tuple_ls
        ]
        return bw_keyvalue_str

    @property
    def bw_strv1(self):
        """Return Bandwidth Line string following spec v1.X.X."""
        bw_line_str = (
            BWLINE_KEYVALUES_SEP_V1.join(self.bw_keyvalue_v1str_ls) + LINE_SEP
        )
        if len(bw_line_str) > BW_LINE_SIZE:
            # if this is the case, probably there are too many KeyValues,
            # or the limit needs to be changed in Tor
            log.warn(
                "The bandwidth line %s is longer than %s",
                len(bw_line_str),
                BW_LINE_SIZE,
            )
        return bw_line_str


class V3BWFile(object):
    """
    Create a Bandwidth List file following spec version 1.X.X

    :param V3BWHeader v3bwheader: header
    :param list v3bwlines: V3BWLines
    """

    def __init__(self, v3bwheader, v3bwlines):
        self.header = v3bwheader
        self.bw_lines = v3bwlines

    def __str__(self):
        return str(self.header) + "".join(
            [str(bw_line) or "" for bw_line in self.bw_lines]
        )

    @classmethod
    def from_results(
        cls,
        results,
        scanner_country=None,
        destinations_countries=None,
        state_fpath="",
        scale_constant=SBWS_SCALE_CONSTANT,
        scaling_method=TORFLOW_SCALING,
        torflow_obs=TORFLOW_OBS_LAST,
        torflow_cap=TORFLOW_BW_MARGIN,
        round_digs=PROP276_ROUND_DIG,
        secs_recent=None,
        secs_away=None,
        min_num=0,
        consensus_path=None,
        max_bw_diff_perc=MAX_BW_DIFF_PERC,
        reverse=False,
    ):
        """Create V3BWFile class from sbws Results.

        :param dict results: see below
        :param str state_fpath: path to the state file
        :param int scaling_method:
            Scaling method to obtain the bandwidth
            Possible values: {None, SBWS_SCALING, TORFLOW_SCALING} = {0, 1, 2}
        :param int scale_constant: sbws scaling constant
        :param int torflow_obs: method to choose descriptor observed bandwidth
        :param bool reverse: whether to sort the bw lines descending or not

        Results are in the form::

            {'relay_fp1': [Result1, Result2, ...],
             'relay_fp2': [Result1, Result2, ...]}

        """
        log.info("Processing results to generate a bandwidth list file.")
        header = V3BWHeader.from_results(
            results, scanner_country, destinations_countries, state_fpath
        )
        bw_lines_raw = []
        bw_lines_excluded = []
        router_statuses_d = cls.read_router_statuses(consensus_path)
        # XXX: Use router_statuses_d to not parse again the file.
        number_consensus_relays = cls.read_number_consensus_relays(
            consensus_path
        )
        state = State(state_fpath)

        # Create a dictionary with the number of relays excluded by any of the
        # of the filtering rules that makes relays non-`eligible`.
        # NOTE: In HEADER_RECENT_MEASUREMENTS_EXCLUDED_KEYS it is
        # explained what are the KeyValues.
        # See also the comments in `from_results`.
        exclusion_dict = dict(
            [(k, 0) for k in HEADER_RECENT_MEASUREMENTS_EXCLUDED_KEYS]
        )
        for fp, values in results.items():
            # log.debug("Relay fp %s", fp)
            line, reason = V3BWLine.from_results(
                values, secs_recent, secs_away, min_num, router_statuses_d
            )
            # If there is no reason it means the line will not be excluded.
            if not reason:
                bw_lines_raw.append(line)
            else:
                # Store the excluded lines to include them in the bandwidth
                # file.
                bw_lines_excluded.append(line)
                exclusion_dict[reason] = exclusion_dict.get(reason, 0) + 1
        # Add the headers with the number of excluded relays by reason
        header.add_relays_excluded_counters(exclusion_dict)

        if not bw_lines_raw:
            # It could be possible to scale the lines that were successful
            # even if excluded, but is not done here.
            log.info(
                "After applying restrictions to the raw results, "
                "there is not any. Scaling can not be applied."
            )
            # Update the header and log the progress.
            cls.update_progress(cls, 0, header, number_consensus_relays, state)
            # Set the lines that would be excluded anyway (`vote=0`) with
            # `under_min_report=1`
            cls.set_under_min_report(bw_lines_excluded)
            # Create the bandwidth file with the lines that would be excluded.
            return cls(header, bw_lines_excluded)
        if scaling_method == SBWS_SCALING:
            bw_lines = cls.bw_sbws_scale(bw_lines_raw, scale_constant)
            cls.warn_if_not_accurate_enough(bw_lines, scale_constant)
            # log.debug(bw_lines[-1])
        elif scaling_method == TORFLOW_SCALING:
            bw_lines = cls.bw_torflow_scale(
                bw_lines_raw,
                torflow_obs,
                torflow_cap,
                round_digs,
                router_statuses_d=router_statuses_d,
            )
            # log.debug(bw_lines[-1])
            # Update the header and log the progress.
            min_perc = cls.update_progress(
                cls, len(bw_lines), header, number_consensus_relays, state
            )
            # If after scaling the number of lines is less than the percentage
            # of lines to report, set them with `under_min_report`.
            if not min_perc:
                cls.set_under_min_report(bw_lines)
        else:
            bw_lines = cls.bw_kb(bw_lines_raw)
            # log.debug(bw_lines[-1])
        # Not using the result for now, just warning
        cls.is_max_bw_diff_perc_reached(
            bw_lines, max_bw_diff_perc, router_statuses_d
        )
        header.add_time_report_half_network()
        f = cls(header, bw_lines + bw_lines_excluded)
        return f

    @classmethod
    def from_v1_fpath(cls, fpath):
        log.info("Parsing bandwidth file %s", fpath)
        with open(fpath) as fd:
            text = fd.read()
        all_lines = text.split(LINE_SEP)
        header, lines = V3BWHeader.from_lines_v1(all_lines)
        bw_lines = [V3BWLine.from_bw_line_v1(line) for line in lines]
        return cls(header, bw_lines)

    @classmethod
    def from_v100_fpath(cls, fpath):
        log.info("Parsing bandwidth file %s", fpath)
        with open(fpath) as fd:
            text = fd.read()
        all_lines = text.split(LINE_SEP)
        header, lines = V3BWHeader.from_lines_v100(all_lines)
        bw_lines = sorted(
            [V3BWLine.from_bw_line_v1(l) for l in lines], key=lambda l: l.bw
        )
        return cls(header, bw_lines)

    @staticmethod
    def set_under_min_report(bw_lines):
        """
        Mondify the Bandwidth Lines adding the KeyValue `under_min_report`,
        `vote`.
        """
        log.debug("Setting `under_min_report` to %s lines.", len(bw_lines))
        for l in bw_lines:
            l.under_min_report = 1
            l.vote = 0

    @staticmethod
    def bw_kb(bw_lines, reverse=False):
        bw_lines_scaled = copy.deepcopy(bw_lines)
        for l in bw_lines_scaled:
            l.bw = max(round(l.bw / 1000), 1)
        return sorted(bw_lines_scaled, key=lambda x: x.bw, reverse=reverse)

    @staticmethod
    def bw_sbws_scale(
        bw_lines, scale_constant=SBWS_SCALE_CONSTANT, reverse=False
    ):
        """Return a new V3BwLine list scaled using sbws method.

        :param list bw_lines:
            bw lines to scale, not self.bw_lines,
            since this method will be before self.bw_lines have been
            initialized.
        :param int scale_constant:
            the constant to multiply by the ratio and
            the bandwidth to obtain the new bandwidth
        :returns list: V3BwLine list
        """
        log.debug("Scaling bandwidth using sbws method.")
        m = median([l.bw for l in bw_lines])
        bw_lines_scaled = copy.deepcopy(bw_lines)
        for l in bw_lines_scaled:
            # min is to limit the bw to descriptor average-bandwidth
            # max to avoid bandwidth with 0 value
            l.bw = max(
                round(min(l.desc_bw_avg, l.bw * scale_constant / m) / 1000), 1
            )
        return sorted(bw_lines_scaled, key=lambda x: x.bw, reverse=reverse)

    @staticmethod
    def warn_if_not_accurate_enough(
        bw_lines, scale_constant=SBWS_SCALE_CONSTANT
    ):
        margin = 0.001
        accuracy_ratio = median([l.bw for l in bw_lines]) / scale_constant
        log.info(
            "The generated lines are within {:.5}% of what they should "
            "be".format((1 - accuracy_ratio) * 100)
        )
        if accuracy_ratio < 1 - margin or accuracy_ratio > 1 + margin:
            log.warning(
                "There was %f%% error and only +/- %f%% is " "allowed",
                (1 - accuracy_ratio) * 100,
                margin * 100,
            )

    @staticmethod
    def is_max_bw_diff_perc_reached(
        bw_lines, max_bw_diff_perc=MAX_BW_DIFF_PERC, router_statuses_d=None
    ):
        if router_statuses_d:
            sum_consensus_bw = sum(
                list(
                    map(
                        lambda x: x.bandwidth * 1000,
                        router_statuses_d.values(),
                    )
                )
            )
        else:
            sum_consensus_bw = sum(
                [
                    l.consensus_bandwidth
                    for l in bw_lines
                    if getattr(l, "consensus_bandwidth", None)
                ]
            )
        # Because the scaled bandwidth is in KB, but not the stored consensus
        # bandwidth, multiply by 1000.
        # Do not count the bandwidths for the relays that were excluded
        sum_bw = sum([l.bw for l in bw_lines if getattr(l, "vote", 1)]) * 1000
        # Percentage difference
        diff_perc = (
            abs(sum_consensus_bw - sum_bw)
            # Avoid ZeroDivisionError
            / (max(1, (sum_consensus_bw + sum_bw)) / 2)
        ) * 100
        log.info(
            "The difference between the sum of the last consensus weight (%s "
            "KB) and the sum of the reported weight in this Bandwidth File "
            "(%s KB) is %s%%.",
            sum_consensus_bw,
            sum_bw,
            round(diff_perc),
        )
        if diff_perc > max_bw_diff_perc:
            log.warning(
                "The difference between the sum of the last consensus weight "
                "(%s KB) and the sum of the reported weight in this Bandwidth "
                "File (%s KB) is %s%%. It is more than %s%%.",
                sum_consensus_bw,
                sum_bw,
                round(diff_perc),
                max_bw_diff_perc,
            )
            return True
        return False

    @staticmethod
    def bw_torflow_scale(
        bw_lines,
        desc_bw_obs_type=TORFLOW_OBS_MEAN,
        cap=TORFLOW_BW_MARGIN,
        num_round_dig=PROP276_ROUND_DIG,
        reverse=False,
        router_statuses_d=None,
    ):
        """
        Obtain final bandwidth measurements applying Torflow's scaling
        method.

        See details in :ref:`torflow_aggr`.
        """
        log.info("Calculating relays' bandwidth using Torflow method.")
        bw_lines_tf = copy.deepcopy(bw_lines)
        # mean (Torflow's strm_avg)
        mu = mean([l.bw_mean for l in bw_lines])
        # filtered mean (Torflow's filt_avg)
        muf = mean([l.bw_filt for l in bw_lines])
        log.debug("mu %i bytes.", mu)
        log.debug("muf %i bytes.", muf)

        # Torflow's ``tot_net_bw``, sum of the scaled bandwidth for the relays
        # that are in the last consensus
        sum_bw = 0
        for l in bw_lines_tf:
            # First, obtain the observed bandwidth, later check what to do
            # if it is 0 or None.
            if desc_bw_obs_type == TORFLOW_OBS_LAST:
                # In case there's no last, use the mean, because it is possible
                # that it went down for a few days, but no more than 5,
                # otherwise the mean will be 1
                desc_bw_obs = l.desc_bw_obs_last or l.desc_bw_obs_mean
            # Assume that if it is not TORFLOW_OBS_LAST, then it is
            # TORFLOW_OBS_MEAN
            else:
                desc_bw_obs = l.desc_bw_obs_mean

            # Excerpt from bandwidth-file-spec.txt section 2.3
            # A relay's MaxAdvertisedBandwidth limits the bandwidth-avg in its
            # descriptor.
            # Therefore generators MUST limit a relay's measured bandwidth to
            # its descriptor's bandwidth-avg.
            # Generators SHOULD NOT limit measured bandwidths based on
            # descriptors' bandwidth-observed, because that penalises new
            # relays.
            # In the case that descriptor average or burst are None,
            # ignore them since it must be a bug in ``Resultdump``, already
            # logged in x_bw/bandwidth_x_from_results, but scale.
            if desc_bw_obs is not None:
                if l.desc_bw_bur is not None:
                    if l.desc_bw_avg is not None:
                        desc_bw = min(
                            desc_bw_obs, l.desc_bw_bur, l.desc_bw_avg
                        )
                    else:
                        desc_bw = min(desc_bw_obs, l.desc_bw_bur)
                else:
                    if l.desc_bw_avg is not None:
                        desc_bw = min(desc_bw_obs, l.desc_bw_avg)
                    else:
                        desc_bw = desc_bw_obs
            else:
                log.warning("Can not scale relay missing descriptor.")
                continue

            # Torflow's scaling
            ratio_stream = l.bw_mean / mu
            ratio_stream_filtered = l.bw_filt / muf
            ratio = max(ratio_stream, ratio_stream_filtered)

            # Assign it to an attribute, so it's not lost before capping and
            # rounding
            l.bw = ratio * desc_bw

            # If the consensus is available, sum only the bw for the relays
            # that are in the consensus
            if router_statuses_d:
                if l.node_id.replace("$", "") in router_statuses_d:
                    sum_bw += l.bw
            # Otherwise sum all bw, for compatibility with tests that were not
            # using the consensus file.
            else:
                sum_bw += l.bw

        # Cap maximum bw, only possible when the ``sum_bw`` is calculated.
        # Torflow's clipping
        hlimit = sum_bw * cap
        log.debug(
            "Sum of all the reported relays' scaled bandwidth: %i bytes, "
            "the limit for any relay is: %i bytes",
            sum_bw,
            hlimit,
        )
        for l in bw_lines_tf:
            bw_scaled = min(hlimit, l.bw)
            # round and convert to KB
            l.bw = kb_round_x_sig_dig(bw_scaled, digits=num_round_dig)
        return sorted(bw_lines_tf, key=lambda x: x.bw, reverse=reverse)

    @staticmethod
    def read_number_consensus_relays(consensus_path):
        """Read the number of relays in the Network from the cached consensus
        file."""
        num = None
        try:
            num = len(list(parse_file(consensus_path)))
        except (FileNotFoundError, AttributeError):
            log.info(
                "It is not possible to obtain statistics about the "
                "percentage of measured relays because the cached "
                "consensus file is not found."
            )
        log.debug("Number of relays in the network %s", num)
        return num

    @staticmethod
    def read_router_statuses(consensus_path):
        """Read the router statuses from the cached consensus file."""
        router_statuses_d = None
        try:
            router_statuses_d = dict(
                [(r.fingerprint, r) for r in parse_file(consensus_path)]
            )
        except (FileNotFoundError, AttributeError):
            log.warning(
                "It is not possible to obtain the last consensus "
                "cached file %s.",
                consensus_path,
            )
        return router_statuses_d

    @staticmethod
    def measured_progress_stats(
        num_bw_lines, number_consensus_relays, min_perc_reached_before
    ):
        """Statistics about measurements progress,
        to be included in the header.

        :param list bw_lines: the bw_lines after scaling and applying filters.
        :param str consensus_path: the path to the cached consensus file.
        :param str state_fpath: the path to the state file
        :returns dict, bool: Statistics about the progress made with
            measurements and whether the percentage of measured relays has been
            reached.

        """
        # cached-consensus should be updated every time that scanner get the
        # network status or descriptors?
        # It will not be updated to the last consensus, but the list of
        # measured relays is not either.
        assert isinstance(number_consensus_relays, int)
        assert isinstance(num_bw_lines, int)
        statsd = {}
        statsd["number_eligible_relays"] = num_bw_lines
        statsd["number_consensus_relays"] = number_consensus_relays
        statsd["minimum_number_eligible_relays"] = round(
            statsd["number_consensus_relays"] * MIN_REPORT / 100
        )
        statsd["percent_eligible_relays"] = round(
            num_bw_lines * 100 / statsd["number_consensus_relays"]
        )
        statsd["minimum_percent_eligible_relays"] = MIN_REPORT
        if (
            statsd["number_eligible_relays"]
            < statsd["minimum_number_eligible_relays"]
        ):
            # if min percent was was reached before, warn
            # otherwise, debug
            if min_perc_reached_before is not None:
                log.warning(
                    "The percentage of the measured relays is less "
                    "than the %s%% of the relays in the network (%s).",
                    MIN_REPORT,
                    statsd["number_consensus_relays"],
                )
            else:
                log.info(
                    "The percentage of the measured relays is less "
                    "than the %s%% of the relays in the network (%s).",
                    MIN_REPORT,
                    statsd["number_consensus_relays"],
                )
            return statsd, False
        return statsd, True

    @property
    def is_min_perc(self):
        if getattr(self.header, "number_eligible_relays", 0) < getattr(
            self.header, "minimum_number_eligible_relays", 0
        ):
            return False
        return True

    @property
    def sum_bw(self):
        return sum([l.bw for l in self.bw_lines if hasattr(l, "bw")])

    @property
    def num(self):
        return len(self.bw_lines)

    @property
    def mean_bw(self):
        return mean([l.bw for l in self.bw_lines if hasattr(l, "bw")])

    @property
    def median_bw(self):
        return median([l.bw for l in self.bw_lines if hasattr(l, "bw")])

    @property
    def max_bw(self):
        return max([l.bw for l in self.bw_lines if hasattr(l, "bw")])

    @property
    def min_bw(self):
        return min([l.bw for l in self.bw_lines if hasattr(l, "bw")])

    @property
    def info_stats(self):
        if not self.bw_lines:
            return
        log.info("Bandwidth File with:")
        for attr in [
            "sum_bw",
            "mean_bw",
            "median_bw",
            "max_bw",
            "min_bw",
        ]:
            log.info("%s: %s KB.", attr, str(getattr(self, attr)))
        log.info("%s Bandwidth Lines", str(getattr(self, "num")))

    def update_progress(
        self, num_bw_lines, header, number_consensus_relays, state
    ):
        """
        Returns True if the minimim percent of Bandwidth Lines was reached
        and False otherwise.
        Update the header with the progress.
        """
        min_perc_reached_before = state.get("min_perc_reached")
        if number_consensus_relays is not None:
            statsd, success = self.measured_progress_stats(
                num_bw_lines, number_consensus_relays, min_perc_reached_before
            )
            # add statistics about progress always
            header.add_stats(**statsd)
            if not success:
                # From sbws 1.1.0 the lines are reported (#29853) even if they
                # are less than the minimum percent.
                state["min_perc_reached"] = None
                return False
            else:
                state["min_perc_reached"] = now_isodt_str()
                return True

    def bw_line_for_node_id(self, node_id):
        """Returns the bandwidth line for a given node fingerprint.

        Used to combine data when plotting.
        """
        bwl = [l for l in self.bw_lines if l.node_id == node_id]
        if bwl:
            return bwl[0]
        return None

    def to_plt(self, attrs=["bw"], sorted_by=None):
        """Return bandwidth data in a format useful for matplotlib.

        Used from external tool to plot.
        """
        x = [i for i in range(0, self.num)]
        ys = [[getattr(l, k) for l in self.bw_lines] for k in attrs]
        return x, ys, attrs

    def write(self, output):
        if output == "/dev/stdout":
            log.info("Writing to stdout is not supported.")
            return
        log.info("Writing v3bw file to %s", output)
        # To avoid inconsistent reads, the bandwidth data is written to an
        # archive path, then atomically symlinked to 'latest.v3bw'
        out_dir = os.path.dirname(output)
        out_link = os.path.join(out_dir, "latest.v3bw")
        out_link_tmp = out_link + ".tmp"
        with DirectoryLock(out_dir):
            with open(output, "wt") as fd:
                fd.write(str(self.header))
                for line in self.bw_lines:
                    fd.write(str(line))
            output_basename = os.path.basename(output)
            # To atomically symlink a file, we need to create a temporary link,
            # then rename it to the final link name. (POSIX guarantees that
            # rename is atomic.)
            log.debug(
                "Creating symlink {} -> {}.".format(
                    out_link_tmp, output_basename
                )
            )
            os.symlink(output_basename, out_link_tmp)
            log.debug(
                "Renaming symlink {} -> {} to {} -> {}.".format(
                    out_link_tmp, output_basename, out_link, output_basename
                )
            )
            os.rename(out_link_tmp, out_link)
